"""
================================================
04. From MEG sensor-space data to HNN simulation
================================================

This example demonstrates how to calculate an inverse solution of the median
nerve evoked response potential (ERP) in S1 from the MNE somatosensory dataset,
and then simulate a biophysical model network that reproduces the observed
dynamics. Note that we do not expound on how we came up with the sequence of
evoked drives used in this example, rather, we only demonstrate its
implementation. For those who want more background on the HNN model and the
process used to articulate the proximal and distal drives needed to simulate
evoked responses, see the `HNN ERP tutorial`_. The sequence of evoked drives
presented here is not part of a current publication but is motivated by prior
studies [1]_, [2]_.
"""

# Authors: Mainak Jas <mainakjas@gmail.com>
#          Ryan Thorpe <ryan_thorpe@brown.edu>

# sphinx_gallery_thumbnail_number = 2

###############################################################################
# First, we will import the packages needed for computing the inverse solution
# from the MNE somatosensory dataset. `MNE`_ (and its dependency `NiBabel`_)
# can be installed with ``pip install mne nibabel``
# The somatosensory dataset can be downloaded by
# importing ``somato`` from ``mne.datasets``.
import os.path as op
import matplotlib.pyplot as plt

import mne
from mne.datasets import somato
from mne.minimum_norm import apply_inverse, make_inverse_operator

###############################################################################
# Now we set the the path of the ``somato`` dataset for subject ``'01'``.
data_path = somato.data_path()
subject = '01'
task = 'somato'
raw_fname = op.join(data_path, 'sub-{}'.format(subject), 'meg',
                    'sub-{}_task-{}_meg.fif'.format(subject, task))
fwd_fname = op.join(data_path, 'derivatives', 'sub-{}'.format(subject),
                    'sub-{}_task-{}-fwd.fif'.format(subject, task))
subjects_dir = op.join(data_path, 'derivatives', 'freesurfer', 'subjects')

###############################################################################
# Then, we load the raw data and estimate the inverse operator.

# Read and band-pass filter the raw data
raw = mne.io.read_raw_fif(raw_fname, preload=True)
l_freq, h_freq = 1, 40
raw.filter(l_freq, h_freq)

# Identify stimulus events associated with MEG time series in the dataset
events = mne.find_events(raw, stim_channel='STI 014')

# Define epochs within the time series
event_id, tmin, tmax = 1, -.2, .17
baseline = None
epochs = mne.Epochs(raw, events, event_id, tmin, tmax, baseline=baseline,
                    reject=dict(grad=4000e-13, eog=350e-6), preload=True)

# Compute the inverse operator
fwd = mne.read_forward_solution(fwd_fname)
cov = mne.compute_covariance(epochs)
inv = make_inverse_operator(epochs.info, fwd, cov)

###############################################################################
# There are several methods to do source reconstruction. Some of the methods
# such as MNE are distributed source methods whereas dipole fitting will
# estimate the location and amplitude of a single current dipole. At the
# moment, we do not offer explicit recommendations on which source
# reconstruction technique is best for HNN. However, we do want our users
# to note that the dipole currents simulated with HNN are assumed to be normal
# to the cortical surface. Hence, using the option ``pick_ori='normal'`` is
# appropriate.
snr = 3.
lambda2 = 1. / snr ** 2
evoked = epochs.average()
stc = apply_inverse(evoked, inv, lambda2, method='MNE',
                    pick_ori="normal", return_residual=False,
                    verbose=True)

###############################################################################
# To extract the primary response in primary somatosensory cortex (S1), we
# create a label for the postcentral gyrus (S1) in source-space
hemi = 'rh'
label_tag = 'G_postcentral'
label_s1 = mne.read_labels_from_annot(subject, parc='aparc.a2009s', hemi=hemi,
                                      regexp=label_tag,
                                      subjects_dir=subjects_dir)[0]

###############################################################################
# Visualizing the distributed S1 activation in reference to the geometric
# structure of the cortex (i.e., plotted on a structural MRI) can help us
# figure out how to orient the dipole. Note that in the HNN framework,
# positive and negative deflections of a current dipole source correspond to
# upwards (from deep to superficial) and downwards (from superficial to deep)
# current flow, respectively. Uncomment the following code to open an
# interactive 3D render of the brain and its surface activation (requires the
# ``pyvista`` python library). You should get 2 plots, the first showing the
# post-central gyrus label from which the dipole time course was extracted and
# the second showing MNE activation at 0.040 sec that resemble the following
# images.
'''
Brain = mne.viz.get_brain_class()
brain_label = Brain(subject, hemi, 'white', subjects_dir=subjects_dir)
brain_label.add_label(label_s1, color='green', alpha=0.9)
stc_label = stc.in_label(label_s1)
brain = stc_label.plot(subjects_dir=subjects_dir, hemi=hemi, surface='white',
                       view_layout='horizontal', initial_time=0.04,
                       backend='pyvista')
'''

###############################################################################
# |mne_label_fig|
#
# |mne_activity_fig|

###############################################################################
# Now we extract the representative time course of dipole activation in our
# labeled brain region using ``mode='pca_flip'`` (see `this MNE-python
# example`_ for more details). Note that the most prominent component of the
# median nerve response occurs in the posterior wall of the central sulcus at
# ~0.040 sec. Since the dipolar activity here is negative, we orient the
# extracted waveform so that the deflection at ~0.040 sec is pointed downwards.
# Thus, the ~0.040 sec deflection corresponds to current flow traveling from
# superficial to deep layers of cortex.
flip_data = stc.extract_label_time_course(label_s1, inv['src'],
                                          mode='pca_flip')
dipole_tc = -flip_data[0] * 1e9

plt.figure()
plt.plot(1e3 * stc.times, dipole_tc, 'ro--')
plt.xlabel('Time (ms)')
plt.ylabel('Current Dipole (nAm)')
plt.xlim((0, 170))
plt.axhline(0, c='k', ls=':')
plt.show()

###############################################################################
# Now, let us try to simulate the same with ``hnn-core``. We read in the
# network parameters from ``N20.json`` and instantiate the network.

import hnn_core
from hnn_core import simulate_dipole, jones_2009_model
from hnn_core import average_dipoles, JoblibBackend

hnn_core_root = op.dirname(hnn_core.__file__)
params_fname = op.join(hnn_core_root, 'param', 'N20.json')
net = jones_2009_model(params_fname)

###############################################################################
# To simulate the source of the median nerve evoked response, we add a
# sequence of synchronous evoked drives: 1 proximal, 2 distal, and 1 final
# proximal drive. In order to understand the physiological implications of
# proximal and distal drive as well as the general process used to articulate
# a sequence of exogenous drive for simulating evoked responses, see the
# `HNN ERP tutorial`_. Note that setting ``n_drive_cells=1`` and
# ``cell_specific=False`` creates a drive with synchronous input across cells
# in the network.

# Early proximal drive
weights_ampa_p = {'L2_basket': 0.0036, 'L2_pyramidal': 0.0039,
                  'L5_basket': 0.0019, 'L5_pyramidal': 0.0020}
weights_nmda_p = {'L2_basket': 0.0029, 'L2_pyramidal': 0.0005,
                  'L5_basket': 0.0030, 'L5_pyramidal': 0.0019}
synaptic_delays_p = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_basket': 1.0, 'L5_pyramidal': 1.0}

net.add_evoked_drive(
    'evprox1', mu=21., sigma=4., numspikes=1, location='proximal',
    n_drive_cells=1, cell_specific=False, weights_ampa=weights_ampa_p,
    weights_nmda=weights_nmda_p, synaptic_delays=synaptic_delays_p,
    event_seed=276)

# Late proximal drive
weights_ampa_p = {'L2_basket': 0.003, 'L2_pyramidal': 0.0039,
                  'L5_basket': 0.004, 'L5_pyramidal': 0.0020}
weights_nmda_p = {'L2_basket': 0.001, 'L2_pyramidal': 0.0005,
                  'L5_basket': 0.002, 'L5_pyramidal': 0.0020}
synaptic_delays_p = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_basket': 1.0, 'L5_pyramidal': 1.0}

net.add_evoked_drive(
    'evprox2', mu=134., sigma=4.5, numspikes=1, location='proximal',
    n_drive_cells=1, cell_specific=False, weights_ampa=weights_ampa_p,
    weights_nmda=weights_nmda_p, synaptic_delays=synaptic_delays_p,
    event_seed=276)

# Early distal drive
weights_ampa_d = {'L2_basket': 0.0043, 'L2_pyramidal': 0.0032,
                  'L5_pyramidal': 0.0009}
weights_nmda_d = {'L2_basket': 0.0029, 'L2_pyramidal': 0.0051,
                  'L5_pyramidal': 0.0010}
synaptic_delays_d = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_pyramidal': 0.1}

net.add_evoked_drive(
    'evdist1', mu=32., sigma=2.5, numspikes=1, location='distal',
    n_drive_cells=1, cell_specific=False, weights_ampa=weights_ampa_d,
    weights_nmda=weights_nmda_d, synaptic_delays=synaptic_delays_d,
    event_seed=277)

# Late distal drive
weights_ampa_d = {'L2_basket': 0.0041, 'L2_pyramidal': 0.0019,
                  'L5_pyramidal': 0.0018}
weights_nmda_d = {'L2_basket': 0.0032, 'L2_pyramidal': 0.0018,
                  'L5_pyramidal': 0.0017}
synaptic_delays_d = {'L2_basket': 0.1, 'L2_pyramidal': 0.1,
                     'L5_pyramidal': 0.1}

net.add_evoked_drive(
    'evdist2', mu=84., sigma=4.5, numspikes=1, location='distal',
    n_drive_cells=1, cell_specific=False, weights_ampa=weights_ampa_d,
    weights_nmda=weights_nmda_d, synaptic_delays=synaptic_delays_d,
    event_seed=275)

###############################################################################
# Now we run the simulation over 2 trials so that we can plot the average
# aggregate dipole. For a better match to the empirical waveform, set
# ``n_trials`` to be >=25.
n_trials = 2
# n_trials = 25
with JoblibBackend(n_jobs=2):
    dpls = simulate_dipole(net, tstop=170., n_trials=n_trials)

###############################################################################
# Since the model is a reduced representation of the larger network
# contributing to the response, the model response is noisier than it would be
# in the net activity from a larger network where these effects are averaged
# out, and the dipole amplitude is smaller than the recorded data. The
# post-processing steps of smoothing and scaling the simulated dipole response
# allow us to more accurately approximate the true signal responsible for the
# recorded macroscopic evoked response [1]_, [2]_.
dpl_smooth_win = 20
dpl_scalefctr = 12
for dpl in dpls:
    dpl.smooth(dpl_smooth_win)
    dpl.scale(dpl_scalefctr)

###############################################################################
# Finally, we plot the driving spike histogram, empirical and simulated median
# nerve evoked response waveforms, and output spike histogram.
fig, axes = plt.subplots(3, 1, sharex=True, figsize=(6, 6),
                         constrained_layout=True)
net.cell_response.plot_spikes_hist(ax=axes[0],
                                   spike_types=['evprox', 'evdist'],
                                   show=False)
axes[1].axhline(0, c='k', ls=':', label='_nolegend_')
axes[1].plot(1e3 * stc.times, dipole_tc, 'r--')
average_dipoles(dpls).plot(ax=axes[1], show=False)
axes[1].legend(['MNE label average', 'HNN simulation'])
axes[1].set_ylabel('Current Dipole (nAm)')
net.cell_response.plot_spikes_raster(ax=axes[2])

###############################################################################
# References
# ----------
# .. [1] Jones, S. R., Pritchett, D. L., Stufflebeam, S. M., Hämäläinen, M.
#    & Moore, C. I. Neural correlates of tactile detection: a combined
#    magnetoencephalography and biophysically based computational modeling
#    study. J. Neurosci. 27, 10751–10764 (2007).
# .. [2] Neymotin SA, Daniels DS, Caldwell B, McDougal RA, Carnevale NT,
#    Jas M, Moore CI, Hines ML, Hämäläinen M, Jones SR. Human Neocortical
#    Neurosolver (HNN), a new software tool for interpreting the cellular and
#    network origin of human MEG/EEG data. eLife 9, e51214 (2020).
#    https://doi.org/10.7554/eLife.51214

###############################################################################
# .. LINKS
#
# .. _MNE: https://mne.tools/
# .. _NiBabel: https://nipy.org/nibabel/
# .. _HNN ERP tutorial: https://jonescompneurolab.github.io/hnn-tutorials/erp/erp
# .. _this MNE-python example: https://mne.tools/stable/auto_examples/inverse/plot_label_source_activations.html
# .. |mne_label_fig| image:: https://user-images.githubusercontent.com/20212206/106524603-cfe75c80-64b0-11eb-9607-3415195c3e7a.png
#   :width: 400
# .. |mne_activity_fig| image:: https://user-images.githubusercontent.com/20212206/106524542-b514e800-64b0-11eb-835e-497454e75eb9.png
#   :width: 400



"""Bayesian optimization according to:

Brochu, Cora, and de Freitas' tutorial at
http://haikufactory.com/files/bayopt.pdf

Adopted from http://atpassos.me/post/44900091837/bayesian-optimization
"""

# Authors: Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Alexandre Passos <alexandre.tp@gmail.com>
#          Mainak Jas <mjas@mgh.harvard.edu>
#          Carolina Fernandez <cxf418@miami.edu>

import warnings
from sklearn import gaussian_process
import numpy as np

import scipy.stats as st


def expected_improvement(gp, best_f, all_x):
    """The expected improvement acquisition function. The equation is
    explained in Eq (3) of the tutorial.

    Parameters
    ----------
    gp : instance of GaussianProcessRegressor
        The GaussianProcessRegressor object.
    best_f : float
        Best objective value.
    all_x : ndarray
        Randomly distributed samples.

    Returns
    -------
    ndarray
        Object values corresponding to all_x.
    """

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # (n_samples, n_features)
        y, y_std = gp.predict(all_x, return_std=True)
    Z = (y - best_f) / (y_std + 1e-12)
    return (y - best_f) * st.norm.cdf(Z) + y_std * st.norm.pdf(Z)


def bayes_opt(func, x0, cons, acquisition, maxfun=200, debug=False, random_state=None):
    """The actual bayesian optimization function.

    Parameters
    ----------
    func : func
        The objective function.
    x0 : list
       Initial parameters.
    cons : list of tuples
        Parameter constraints in solver-specific format.
    acquisition : func
        Acquisition function we want to use to find query points.
    maxfun : int, optional
        Maximum number of function evaluations. The default is 200.
    debug : bool, optional
        The default is False.
    random_state : int, optional
        Random state of the GaussianProcessRegressor. The default is None.

    Returns
    -------
    best_x : list
        The argument that minimizes f.
    best_f : float
        The final objective value.
    """

    X, y = list(), list()

    # evaluate
    initial_f = func(x0)
    X.append(x0)
    y.append(initial_f)

    best_x = X[np.argmin(y)]
    best_f = y[np.argmin(y)]
    gp = gaussian_process.GaussianProcessRegressor(random_state=random_state)

    if debug:
        print("iter", -1, "best_x", best_x, best_f)

    for i in range(maxfun):
        # draw samples from distribution
        all_x = np.random.uniform(
            low=[idx[0] for idx in cons],
            high=[idx[1] for idx in cons],
            size=(10000, len(cons)),
        )

        gp.fit(np.array(X), np.array(y))  # (n_samples, n_features)

        # get new set of params
        new_x = all_x[acquisition(gp, best_f, all_x).argmin()]  # lowest obj
        new_x = new_x.tolist()

        # evaluate
        new_f = func(new_x)

        X.append(new_x)
        y.append(new_f)
        if new_f < best_f:
            best_f = new_f
            best_x = new_x

        if debug:
            print("iter", i, "best_x", best_x, best_f)

    return best_x, best_f


if __name__ == "__main__":
    from scipy.optimize import rosen

    opt_params, obj_vals = bayes_opt(
        rosen,
        [0.5, 0.6],
        [(-1, 1), (-1, 1)],
        expected_improvement,
        maxfun=200,
        random_state=1,
    )

    x = np.linspace(-1, 1, 50)
    y = np.linspace(-1, 1, 50)
    X, Y = np.meshgrid(x, y)

    objs = rosen(np.vstack((X.ravel(), Y.ravel())))



# Authors: George Dang <george_dang@brown.edu>
#          Rajat Partani <rajatpartani@gmail.com>

import json
import os.path as op
from pathlib import Path
from time import sleep
from urllib.request import urlretrieve

import numpy as np
import pytest

from hnn_core import (
    simulate_dipole,
    read_params,
    calcium_model,
)

from hnn_core.hnn_io import (
    _cell_response_to_dict,
    _rec_array_to_dict,
    _external_drive_to_dict,
    _str_to_node,
    _conn_to_dict,
    _order_drives,
    read_network_configuration,
)

from regenerate_test_network import jones_2009_additional_features

hnn_core_root = Path(__file__).parents[1]
assets_path = Path(hnn_core_root, "tests", "assets")


@pytest.fixture
def params():
    params_path = Path(hnn_core_root, "param", "default.json")
    params = read_params(params_path)
    params["celsius"] = 37.0
    params["threshold"] = 0.0

    return params


@pytest.fixture
def jones_2009_network():
    # This allows us to define this test network once, but use it as both a
    # fixture here in this file, or regenerate the network itself if used
    # elsewhere.
    net = jones_2009_additional_features()

    return net


@pytest.fixture
def calcium_network(params):
    # Instantiating network along with drives
    net = calcium_model(params=params, add_drives_from_params=True, mesh_shape=(3, 3))

    # Adding bias
    tonic_bias = {"L2_pyramidal": 1.0}
    net.add_tonic_bias(amplitude=tonic_bias)

    # Adding electrode arrays
    electrode_pos = (1, 2, 3)
    net.add_electrode_array("el1", electrode_pos)
    electrode_pos = [(1, 2, 3), (-1, -2, -3)]
    net.add_electrode_array("arr1", electrode_pos)

    return net


def test_eq(jones_2009_network, calcium_network):
    net1 = jones_2009_network
    net2 = calcium_network

    # Check eq of same network
    assert net1 == net1
    # Check eq of different networks
    assert not net1 == net2

    # Check change in drives
    net1_clear_drive = net1.copy()
    net1_clear_drive.clear_drives()
    assert net1_clear_drive != net1

    # Hardwired change in drive attribute
    net1_hard_change_drive = net1.copy()
    net1_hard_change_drive.external_drives["type"] = ""
    assert net1_hard_change_drive != net1

    # Hardwired change in drive weights
    net1_hard_change_drive = net1.copy()
    (net1_hard_change_drive.external_drives["evdist1"]["weights_ampa"]["L2_basket"]) = 0
    assert net1_hard_change_drive != net1


def test_eq_conn(jones_2009_network):
    net1 = jones_2009_network

    # Check a change in connectivity
    net1_clear_conn = net1.copy()
    net1_clear_conn.clear_connectivity()
    assert net1_clear_conn != net1

    # Hardwired change in connectivity attribute
    net1_hard_change_conn = net1.copy()
    net1_hard_change_conn.connectivity[0]["gid_pairs"] = {}
    assert net1_hard_change_conn != net1

    # Hardwired change in connectivity nc_dict
    net1_hard_change_conn = net1.copy()
    net1_hard_change_conn.connectivity[0]["nc_dict"]["A_weight"] = 0
    assert net1_hard_change_conn != net1

    # Check edge case, same number of connections, different replicate in conn
    net1_alt_conn1 = net1.copy()
    net1_alt_conn2 = net1.copy()
    l_conn = net1_alt_conn1.connectivity
    l_conn_rep_start = [l_conn[0]] + l_conn
    l_conn_rep_end = l_conn + [l_conn[-1]]
    net1_alt_conn1.connectivity = l_conn_rep_start
    net1_alt_conn2.connectivity = l_conn_rep_end
    assert net1 != net1_alt_conn1
    assert net1_alt_conn1 == net1_alt_conn1
    assert net1_alt_conn1 != net1_alt_conn2


def test_write_configuration(tmp_path, jones_2009_network):
    """Tests that a json file is written"""

    net = jones_2009_network.copy()
    simulate_dipole(net, tstop=2, n_trials=1, dt=0.5)

    # Check no file is already written
    path_out = tmp_path / "net.json"
    assert not path_out.is_file()

    # Write network check
    jones_2009_network.write_configuration(path_out)
    assert path_out.is_file()

    # Overwrite network check
    last_mod_time1 = path_out.stat().st_mtime
    sleep(0.05)
    jones_2009_network.write_configuration(path_out)
    last_mod_time2 = path_out.stat().st_mtime
    assert last_mod_time1 < last_mod_time2

    # No overwrite check
    with pytest.raises(FileExistsError, match="File already exists at path "):
        jones_2009_network.write_configuration(path_out, overwrite=False)

    # Check no outputs were written
    with open(path_out) as file:
        read_in = json.load(file)

    assert not any([bool(val["times"]) for val in read_in["rec_arrays"].values()])
    assert not any([bool(val["voltages"]) for val in read_in["rec_arrays"].values()])
    assert not any([bool(val["events"]) for val in read_in["external_drives"].values()])
    assert read_in["cell_response"] == {}


def test_cell_response_to_dict(jones_2009_network):
    """Tests _cell_response_to_dict function"""
    net = jones_2009_network

    # When a simulation hasn't been run, return an empty dict
    result1 = _cell_response_to_dict(net, write_output=True)
    assert result1 == dict()

    # Check for cell response dictionary after a simulation
    simulate_dipole(net, tstop=2, n_trials=1, dt=0.5)
    assert net.cell_response is not None
    result2 = _cell_response_to_dict(net, write_output=True)
    assert bool(result2) and isinstance(result2, dict)

    # Check for empty dict if kw supplied
    result3 = _cell_response_to_dict(net, write_output=False)
    assert result3 == dict()


def test_rec_array_to_dict(jones_2009_network):
    """Tests _rec_array_to_dict function"""
    net = jones_2009_network

    # Check rec array times and voltages are in dict after simulation
    simulate_dipole(net, tstop=2, n_trials=1, dt=0.5)
    result = _rec_array_to_dict(net.rec_arrays["el1"], write_output=True)
    assert isinstance(result, dict)
    assert all(
        [
            key in result
            for key in [
                "positions",
                "conductivity",
                "method",
                "min_distance",
                "times",
                "voltages",
            ]
        ]
    )
    assert np.array_equal(result["times"], [0.0, 0.5, 1.0, 1.5, 2.0])
    assert result["voltages"].shape == (1, 1, 5)

    # Check values are empty if write_output keyword is false
    result2 = _rec_array_to_dict(net.rec_arrays["el1"], write_output=False)
    assert result2["times"].size == 0
    assert result2["voltages"].size == 0


def test_conn_to_dict(jones_2009_network):
    """Tests _connectivity_to_list_of_dicts function"""
    net = jones_2009_network

    result = _conn_to_dict(net.connectivity[0])
    assert isinstance(result, dict)
    assert result == {
        "target_type": "L2_basket",
        "target_gids": [0, 1, 2],
        "num_targets": 3,
        "src_type": "evdist1",
        "src_gids": [24, 25, 26],
        "num_srcs": 3,
        "gid_pairs": {"24": [0], "25": [1], "26": [2]},
        "loc": "distal",
        "receptor": "ampa",
        "nc_dict": {
            "A_delay": 0.1,
            "A_weight": 0.006562,
            "lamtha": 3.0,
            "threshold": 0.0,
            "gain": 1.0,
        },
        "allow_autapses": 1,
        "probability": 1.0,
    }


def test_external_drive_to_dict(jones_2009_network):
    """Tests _external_drive_to_dict function"""
    net = jones_2009_network

    simulate_dipole(net, tstop=2, n_trials=1, dt=0.5)
    first_key = list(net.external_drives.keys())[0]
    result = _external_drive_to_dict(net.external_drives[first_key], write_output=True)
    assert isinstance(result, dict)
    assert all(
        [
            key in result
            for key in [
                "type",
                "location",
                "n_drive_cells",
                "event_seed",
                "conn_seed",
                "dynamics",
                "events",
                "weights_ampa",
                "weights_nmda",
                "synaptic_delays",
                "probability",
                "name",
                "target_types",
                "cell_specific",
            ]
        ]
    )
    assert len(result["events"][0]) == 21

    result2 = _external_drive_to_dict(
        net.external_drives[first_key], write_output=False
    )
    assert len(result2["events"]) == 0


def test_str_to_node():
    """Creates a tuple (str,int) from string with a comma"""
    result = _str_to_node("cell_name,0")
    assert isinstance(result, tuple)
    assert isinstance(result[0], str)
    assert isinstance(result[1], int)


def test_order_drives(jones_2009_network):
    """Reorders drive dict by ascending range order"""
    drive_names = list(jones_2009_network.external_drives.keys())
    drive_names_alpha = sorted(drive_names)
    drives_reordered = {
        name: jones_2009_network.external_drives for name in drive_names_alpha
    }
    assert list(drives_reordered.keys()) == [
        "alpha_prox",
        "evdist1",
        "evprox1",
        "evprox2",
        "poisson",
    ]

    drives_by_range = _order_drives(jones_2009_network.gid_ranges, drives_reordered)
    assert list(drives_by_range.keys()) == [
        "evdist1",
        "evprox1",
        "evprox2",
        "alpha_prox",
        "poisson",
    ]


def test_read_configuration_json(jones_2009_network):
    """Read-in of a hdf5 file"""
    net = read_network_configuration(Path(assets_path, "jones2009_3x3_drives.json"))
    assert net == jones_2009_network

    # Read without drives
    net_no_drives = read_network_configuration(
        Path(assets_path, "jones2009_3x3_drives.json"), read_drives=False
    )
    # Check there are no external drives
    assert len(net_no_drives.external_drives) == 0
    # Check there are no external drive connections
    connection_src_types = [
        connection["src_type"] for connection in net_no_drives.connectivity
    ]
    assert not any(
        [src_type in net.external_drives.keys() for src_type in connection_src_types]
    )

    # Read without external bias
    net_no_bias = read_network_configuration(
        Path(assets_path, "jones2009_3x3_drives.json"), read_external_biases=False
    )
    assert len(net_no_bias.external_biases) == 0
    assert len(net_no_bias.external_drives) > 0


def test_read_incorrect_format(tmp_path):
    """Test that error raise when the json do not have a Network label."""

    # Checking object type field not exists error
    dummy_data = dict()

    dummy_data["object_type"] = "NotNetwork"
    file_path = tmp_path / "not_net.json"
    with open(file_path, "w") as file:
        json.dump(dummy_data, file)

    with pytest.raises(ValueError, match="The json should encode a Network object."):
        read_network_configuration(file_path)


def test_network_serialization_metadata(jones_2009_network, tmp_path):
    """Test saving and loading a network with the cell_metadata structure."""
    net_original = jones_2009_network
    net_original.add_evoked_drive(
        "evd1",
        mu=5,
        sigma=1,
        numspikes=1,
        location="distal",
        weights_ampa={"L2_pyramidal": 0.1},
    )

    json_path = tmp_path / "net_metadata.json"

    net_original.write_configuration(json_path)

    net_loaded = read_network_configuration(json_path)

    assert net_loaded == net_original

    # checking the nested structure in the loaded network
    assert isinstance(net_loaded.cell_types["L2_pyramidal"], dict)
    assert "cell_object" in net_loaded.cell_types["L2_pyramidal"]
    assert "cell_metadata" in net_loaded.cell_types["L2_pyramidal"]
    assert net_loaded.cell_types["L2_pyramidal"]["cell_metadata"]["layer"] == "2"


def test_read_run_tutorial_json():
    """Test that the first tutorial Network for ERP can be loaded and ran without error."""
    net_url = (
        "https://raw.githubusercontent.com/jonescompneurolab/"
        "hnn-data/refs/heads/main/"
        "network-configurations/ERPYes100Trials.json"
    )
    net_fname = op.join(hnn_core_root, "param", "ERPYes100Trials.json")
    if not op.exists(net_fname):
        urlretrieve(net_url, net_fname)

    # Test that Network can be created without error.
    net = read_network_configuration(net_fname)

    # Test that Network can be simulated without error.
    _ = simulate_dipole(net, tstop=2, n_trials=1, dt=0.5)



import os.path as op
from os import environ
import io
import itertools
from contextlib import redirect_stdout
from threading import Thread, Event
from time import sleep
from urllib.request import urlretrieve

import numpy as np
from numpy import loadtxt
from numpy.testing import assert_array_equal, assert_allclose, assert_raises

import pytest

import hnn_core
from hnn_core import MPIBackend, jones_2009_model, read_params
from hnn_core.dipole import simulate_dipole
from hnn_core.parallel_backends import (
    requires_mpi4py,
    requires_psutil,
    _determine_cores_hwthreading,
)
from hnn_core.network_builder import NetworkBuilder


def _terminate_mpibackend(event, backend):
    # wait for run_subprocess to start MPI proc and put handle on queue
    proc = backend.proc_queue.get()
    # put the proc back in the queue. used by backend.terminate()
    backend.proc_queue.put(proc)

    # give the process a little time to startup
    sleep(0.1)

    # run terminate until it is successful
    while not event.is_set():
        backend.terminate()
        sleep(0.01)


def test_gid_assignment():
    """Test that gids are assigned without overlap across ranks"""

    net = jones_2009_model(add_drives_from_params=False)
    weights_ampa = {"L2_basket": 1.0, "L2_pyramidal": 2.0, "L5_pyramidal": 3.0}
    syn_delays = {"L2_basket": 0.1, "L2_pyramidal": 0.2, "L5_pyramidal": 0.3}

    net.add_bursty_drive(
        "bursty_dist",
        location="distal",
        burst_rate=10,
        weights_ampa=weights_ampa,
        synaptic_delays=syn_delays,
        cell_specific=False,
        n_drive_cells=5,
    )
    net.add_evoked_drive(
        "evoked_prox",
        mu=1.0,
        sigma=1.0,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=syn_delays,
        cell_specific=True,
        n_drive_cells="n_cells",
    )
    net._instantiate_drives(tstop=20, n_trials=2)

    all_gids = list()
    for type_range in net.gid_ranges.values():
        all_gids.extend(list(type_range))
    all_gids.sort()

    n_hosts = 3
    all_gids_instantiated = list()
    for rank in range(n_hosts):
        net_builder = NetworkBuilder(net)
        net_builder._gid_list = list()
        net_builder._gid_assign(rank=rank, n_hosts=n_hosts)
        all_gids_instantiated.extend(net_builder._gid_list)
    all_gids_instantiated.sort()
    assert all_gids_instantiated == sorted(set(all_gids_instantiated))
    assert all_gids == all_gids_instantiated


# The purpose of this incremental mark is to avoid running the full length
# simulation when there are failures in previous (faster) tests. When a test
# in the sequence fails, all subsequent tests will be marked "xfailed" rather
# than skipped.
@pytest.mark.incremental
@pytest.mark.uses_mpi
class TestParallelBackends:
    dpls_reduced_mpi = None
    dpls_reduced_default = None
    dpls_reduced_joblib = None

    def test_run_default(self, run_hnn_core_fixture):
        """Test consistency between default backend simulation and master"""
        global dpls_reduced_default
        dpls_reduced_default, _ = run_hnn_core_fixture(None, reduced=True)
        # test consistency across all parallel backends for multiple trials
        assert_raises(
            AssertionError,
            assert_array_equal,
            dpls_reduced_default[0].data["agg"],
            dpls_reduced_default[1].data["agg"],
        )

    def test_run_joblibbackend(self, run_hnn_core_fixture):
        """Test consistency between joblib backend simulation with master"""
        global dpls_reduced_default, dpls_reduced_joblib

        dpls_reduced_joblib, _ = run_hnn_core_fixture(
            backend="joblib", n_jobs=2, reduced=True
        )

        for trial_idx in range(len(dpls_reduced_default)):
            assert_array_equal(
                dpls_reduced_default[trial_idx].data["agg"],
                dpls_reduced_joblib[trial_idx].data["agg"],
            )

    @requires_mpi4py
    @requires_psutil
    @pytest.mark.parametrize("sensible_default", [False, True])
    def test_detect_cores(self, sensible_default):
        """Test that multiple cores can be detected"""
        [detected_cores_nohw, detected_hwthreading] = _determine_cores_hwthreading(
            use_hwthreading_if_found=False, sensible_default_cores=sensible_default
        )
        assert detected_cores_nohw > 1
        assert isinstance(detected_hwthreading, bool)

        [detected_cores_yeshw, detected_hwthreading] = _determine_cores_hwthreading(
            use_hwthreading_if_found=True, sensible_default_cores=sensible_default
        )
        assert detected_cores_yeshw > 1
        assert isinstance(detected_hwthreading, bool)

        assert detected_cores_yeshw >= detected_cores_nohw

    @requires_mpi4py
    @requires_psutil
    def test_mpi_nprocs(self):
        """Test that MPIBackend can use more than 1 processor"""
        # if only 1 processor is available, then MPIBackend tests will not
        # be valid
        with MPIBackend() as backend:
            assert backend.n_procs > 1

    @requires_mpi4py
    @requires_psutil
    def test_run_mpibackend(self, run_hnn_core_fixture):
        """Test running a MPIBackend on reduced model"""
        global dpls_reduced_default, dpls_reduced_mpi
        dpls_reduced_mpi, _ = run_hnn_core_fixture(backend="mpi", reduced=True)
        for trial_idx in range(len(dpls_reduced_default)):
            # account for rounding error incured during MPI parallelization
            assert_allclose(
                dpls_reduced_default[trial_idx].data["agg"],
                dpls_reduced_mpi[trial_idx].data["agg"],
                rtol=0,
                atol=1e-14,
            )

    @requires_mpi4py
    @requires_psutil
    def test_terminate_mpibackend(self, run_hnn_core_fixture):
        """Test terminating MPIBackend from thread"""
        hnn_core_root = op.dirname(hnn_core.__file__)
        params_fname = op.join(hnn_core_root, "param", "default.json")
        params = read_params(params_fname)
        params.update(
            {"t_evprox_1": 5, "t_evdist_1": 10, "t_evprox_2": 20, "N_trials": 2}
        )
        net = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))

        with MPIBackend() as backend:
            event = Event()
            # start background thread that will kill all MPIBackends
            # until event.set()
            kill_t = Thread(target=_terminate_mpibackend, args=(event, backend))
            # make thread a daemon in case we throw an exception
            # and don't run event.set() so that py.test will
            # not hang before exiting
            kill_t.daemon = True
            kill_t.start()

            with pytest.warns(UserWarning) as record:
                with pytest.raises(
                    RuntimeError, match="MPI simulation failed. Return code: 1"
                ):
                    simulate_dipole(net, tstop=40)

            event.set()
        expected_string = "Child process failed unexpectedly"
        assert expected_string in record[0].message.args[0]

    @requires_mpi4py
    @requires_psutil
    @pytest.mark.parametrize("use_hwthreading_if_found", [True, False])
    def test_run_mpibackend_oversubscribed(self, use_hwthreading_if_found):
        """Test running MPIBackend with oversubscribed number of procs"""
        hnn_core_root = op.dirname(hnn_core.__file__)
        params_fname = op.join(hnn_core_root, "param", "default.json")
        params = read_params(params_fname)
        params.update(
            {"t_evprox_1": 5, "t_evdist_1": 10, "t_evprox_2": 20, "N_trials": 2}
        )
        net = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))

        # Fail state: try running with more procs than cells in the network
        # (will probably oversubscribe too)
        too_many_procs = net._n_cells + 1

        with pytest.raises(
            ValueError, match=("More MPI processes were assigned than there are cells")
        ):
            with MPIBackend(n_procs=too_many_procs) as backend:
                simulate_dipole(net, tstop=40)

        # Force oversubscription and make sure there are always enough cells in
        # the network
        [detected_cores, detected_hwthreading] = _determine_cores_hwthreading(
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=False,
        )

        oversubscribed_procs = detected_cores + 1
        n_grid_1d = int(np.ceil(np.sqrt(oversubscribed_procs)))
        params.update(
            {"t_evprox_1": 5, "t_evdist_1": 10, "t_evprox_2": 20, "N_trials": 2}
        )
        net = jones_2009_model(
            params, add_drives_from_params=True, mesh_shape=(n_grid_1d, n_grid_1d)
        )

        # Case 1 (default): Check that oversubscription turns on if needed, and
        # provides a warning
        override_oversubscribe_option = None
        with pytest.warns(
            UserWarning,
            match=(
                "Number of requested MPI processes exceeds "
                "available cores. Enabling MPI "
                "oversubscription automatically."
            ),
        ):
            with MPIBackend(
                n_procs=oversubscribed_procs,
                use_hwthreading_if_found=use_hwthreading_if_found,
                override_oversubscribe_option=override_oversubscribe_option,
            ) as backend:
                assert backend.n_procs == oversubscribed_procs
                assert "--oversubscribe" in " ".join(backend.mpi_cmd)
                simulate_dipole(net, tstop=40)

        # Case 2: Check that '--oversubscribe' option is passed if
        # oversubscription is forced on. No simulation is run in this case
        # since MacOS CI runners randomly (but not always) fail to run in the
        # following case, even though the same simulation succeeds in Case 1
        # above. This is probably due to MPI instability with the MacOS CI
        # runners (see #992 for an example).
        override_oversubscribe_option = True
        with MPIBackend(
            n_procs=oversubscribed_procs,
            use_hwthreading_if_found=use_hwthreading_if_found,
            override_oversubscribe_option=override_oversubscribe_option,
        ) as backend:
            assert "--oversubscribe" in " ".join(backend.mpi_cmd)

        # Case 3: Check that the simulation fails if oversubscribe is forced
        # off
        override_oversubscribe_option = False
        with MPIBackend(
            n_procs=oversubscribed_procs,
            use_hwthreading_if_found=use_hwthreading_if_found,
            override_oversubscribe_option=override_oversubscribe_option,
        ) as backend:
            assert "--oversubscribe" not in " ".join(backend.mpi_cmd)
            with pytest.raises(
                RuntimeError, match="MPI simulation failed. Return code: 1"
            ):
                simulate_dipole(net, tstop=40)

    @requires_mpi4py
    @requires_psutil
    @pytest.mark.parametrize(
        "use_hwthreading_if_found,sensible_default_cores,override_oversubscribe_option",
        [
            x
            for x in itertools.product(
                [True, False], [True, False], [None, True, False]
            )
        ],
    )
    def test_run_mpibackend_hwthreading(
        self,
        use_hwthreading_if_found,
        sensible_default_cores,
        override_oversubscribe_option,
    ):
        """Test running MPIBackend with oversubscribed number of procs"""
        hnn_core_root = op.dirname(hnn_core.__file__)
        params_fname = op.join(hnn_core_root, "param", "default.json")
        params = read_params(params_fname)
        params.update(
            {"t_evprox_1": 5, "t_evdist_1": 10, "t_evprox_2": 20, "N_trials": 2}
        )
        net = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))

        n_procs = 2
        # Test that the network runs at all
        with MPIBackend(n_procs=n_procs) as backend:
            simulate_dipole(net, tstop=40)

        [_, detected_hwthreading] = _determine_cores_hwthreading(
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=sensible_default_cores,
        )

        # Case 1 (default): Check that hwthreading turns on if needed
        override_hwthreading_option = None

        # Possibly needed to prevent MPIBackend failures to exit processes
        del net
        net = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))
        with MPIBackend(
            n_procs=n_procs,
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=sensible_default_cores,
            override_oversubscribe_option=override_oversubscribe_option,
            override_hwthreading_option=override_hwthreading_option,
        ) as backend:
            if detected_hwthreading:
                assert "--use-hwthread-cpus" in " ".join(backend.mpi_cmd)
            simulate_dipole(net, tstop=40)

        # Case 2: Check that hwthreading turns on if forced. Note that the
        # simulation should NOT be run in this case, since the underlying
        # hardware will NOT necessarily have hardware-threading available.
        override_hwthreading_option = True
        with MPIBackend(
            n_procs=n_procs,
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=sensible_default_cores,
            override_oversubscribe_option=override_oversubscribe_option,
            override_hwthreading_option=override_hwthreading_option,
        ) as backend:
            assert "--use-hwthread-cpus" in " ".join(backend.mpi_cmd)

        # Case 3: Check that hwthreading turns off if forced off.
        override_hwthreading_option = False

        # Possibly needed to prevent MPIBackend failures to exit processes
        del net
        net = jones_2009_model(params, add_drives_from_params=True, mesh_shape=(3, 3))
        with MPIBackend(
            n_procs=n_procs,
            use_hwthreading_if_found=use_hwthreading_if_found,
            sensible_default_cores=sensible_default_cores,
            override_oversubscribe_option=override_oversubscribe_option,
            override_hwthreading_option=override_hwthreading_option,
        ) as backend:
            assert "--use-hwthread-cpus" not in " ".join(backend.mpi_cmd)
            simulate_dipole(net, tstop=40)

    @pytest.mark.parametrize("backend", ["mpi", "joblib"])
    def test_compare_hnn_core(self, run_hnn_core_fixture, backend, n_jobs=1):
        """Test hnn-core does not break."""
        # small snippet of data on data branch for now. To be deleted
        # later. Data branch should have only commit so it does not
        # pollute the history.
        data_url = (
            "https://raw.githubusercontent.com/jonescompneurolab/"
            "hnn-core/test_data/dpl.txt"
        )
        if not op.exists("dpl.txt"):
            urlretrieve(data_url, "dpl.txt")
        dpl_master = loadtxt("dpl.txt")

        dpls, net = run_hnn_core_fixture(backend=backend)
        dpl = dpls[0].smooth(30).scale(3000)

        # write the dipole to a file and compare
        fname = "./dpl2.txt"
        dpl.write(fname)

        dpl_pr = loadtxt(fname)
        assert_array_equal(dpl_pr[:, 2], dpl_master[:, 2])  # L2
        assert_array_equal(dpl_pr[:, 3], dpl_master[:, 3])  # L5

        # Test spike type counts
        spike_type_counts = {}
        for spike_gid in net.cell_response.spike_gids[0]:
            if net.gid_to_type(spike_gid) not in spike_type_counts:
                spike_type_counts[net.gid_to_type(spike_gid)] = 1
            else:
                spike_type_counts[net.gid_to_type(spike_gid)] += 1
        assert "common" not in spike_type_counts
        assert "exgauss" not in spike_type_counts
        assert "extpois" not in spike_type_counts
        assert spike_type_counts == {
            "evprox1": 270,
            "L2_basket": 55,
            "L2_pyramidal": 114,
            "L5_pyramidal": 396,
            "L5_basket": 86,
            "evdist1": 270,
            "evprox2": 270,
        }


# there are no dependencies if this unit tests fails; no need to be in
# class marked incremental
@requires_mpi4py
@requires_psutil
@pytest.mark.uses_mpi
def test_mpi_failure(run_hnn_core_fixture):
    """Test that an MPI failure is handled and messages are printed"""
    # this MPI parameter will cause a MPI job to fail
    environ["OMPI_MCA_btl"] = "self"

    with pytest.warns(UserWarning) as record:
        with io.StringIO() as buf, redirect_stdout(buf):
            with pytest.raises(RuntimeError, match="MPI simulation failed"):
                run_hnn_core_fixture(backend="mpi", reduced=True, postproc=False)
            stdout = buf.getvalue()

    assert "MPI processes are unable to reach each other" in stdout

    expected_string = "Child process failed unexpectedly"
    assert len(record) == 1
    assert record[0].message.args[0] == expected_string

    del environ["OMPI_MCA_btl"]



# Authors: Mainak Jas <mainakjas@gmail.com>
#          Christopher Bailey <bailey.cj@gmail.com>

import pytest
import os.path as op

import numpy as np

import hnn_core
from hnn_core import Network, read_params
from hnn_core.drives import (
    _drive_cell_event_times,
    _get_prng,
    _create_extpois,
    _create_bursty_input,
)
from hnn_core.network import pick_connection
from hnn_core.network_models import jones_2009_model
from hnn_core import simulate_dipole

hnn_core_root = op.dirname(hnn_core.__file__)


@pytest.fixture
def setup_net():
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)
    net = jones_2009_model(params, mesh_shape=(3, 3))

    return net


def test_external_drive_times():
    """Test the different external drives."""

    drive_type = "invalid_drive"
    dynamics = dict(mu=5, sigma=0.5, numspikes=1)
    tstop = 10
    pytest.raises(ValueError, _drive_cell_event_times, "invalid_drive", dynamics, tstop)
    pytest.raises(
        ValueError, _drive_cell_event_times, "ss", dynamics, tstop
    )  # ambiguous

    # validate poisson input time interval
    drive_type = "poisson"
    dynamics = {
        "tstart": 0,
        "tstop": 250.0,
        "rate_constant": {
            "L2_basket": 1,
            "L2_pyramidal": 140.0,
            "L5_basket": 1,
            "L5_pyramidal": 40.0,
        },
    }
    with pytest.raises(ValueError, match="The end time for Poisson input"):
        dynamics["tstop"] = -1
        event_times = _drive_cell_event_times(
            drive_type=drive_type, dynamics=dynamics, tstop=tstop
        )
    with pytest.raises(ValueError, match="The start time for Poisson"):
        dynamics["tstop"] = tstop
        dynamics["tstart"] = -1
        event_times = _drive_cell_event_times(
            drive_type=drive_type, dynamics=dynamics, tstop=tstop
        )

    # checks the poisson spike train generation
    prng = np.random.RandomState()
    lamtha = 50.0
    event_times = _create_extpois(t0=0, T=100000, lamtha=lamtha, prng=prng)
    event_intervals = np.diff(event_times)
    assert pytest.approx(event_intervals.mean(), abs=1.5) == 1000 * 1 / lamtha

    with pytest.raises(ValueError, match="The start time for Poisson"):
        _create_extpois(t0=-5, T=5, lamtha=lamtha, prng=prng)
    with pytest.raises(ValueError, match="The end time for Poisson"):
        _create_extpois(t0=50, T=20, lamtha=lamtha, prng=prng)
    with pytest.raises(ValueError, match="Rate must be > 0"):
        _create_extpois(t0=0, T=1000, lamtha=-5, prng=prng)

    # check bursty/rhythmic input
    t0 = 0
    t0_stdev = 5
    tstop = 100
    f_input = 20.0
    events_per_cycle = 3
    cycle_events_isi = 7
    events_jitter_std = 5.0
    prng, prng2 = _get_prng(seed=0, gid=5)
    event_times = _create_bursty_input(
        t0=t0,
        t0_stdev=t0_stdev,
        tstop=tstop,
        f_input=f_input,
        events_jitter_std=events_jitter_std,
        events_per_cycle=events_per_cycle,
        cycle_events_isi=cycle_events_isi,
        prng=prng,
        prng2=prng2,
    )

    events_per_cycle = 5
    cycle_events_isi = 20
    with pytest.raises(
        ValueError, match=r"(?s)Burst duration .* cannot be greater than"
    ):
        _create_bursty_input(
            t0=t0,
            t0_stdev=t0_stdev,
            tstop=tstop,
            f_input=f_input,
            events_jitter_std=events_jitter_std,
            events_per_cycle=events_per_cycle,
            cycle_events_isi=cycle_events_isi,
            prng=prng,
            prng2=prng2,
        )


def test_drive_seeds(setup_net):
    """Test that unique spike times are generated across trials"""
    net = setup_net
    weights_ampa = {
        "L2_basket": 0.3,
        "L2_pyramidal": 0.3,
        "L5_basket": 0.3,
        "L5_pyramidal": 0.3,
    }
    synaptic_delays = {
        "L2_basket": 0.1,
        "L2_pyramidal": 0.1,
        "L5_basket": 1.0,
        "L5_pyramidal": 1.0,
    }
    net.add_evoked_drive(
        "prox",
        mu=40,
        sigma=8.33,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=synaptic_delays,
        event_seed=1,
    )

    _ = simulate_dipole(net, tstop=100, dt=0.5, n_trials=2)
    trial1_spikes = np.array(sorted(net.external_drives["prox"]["events"][0]))
    trial2_spikes = np.array(sorted(net.external_drives["prox"]["events"][1]))
    # No two spikes should be perfectly identical across seeds
    assert ~np.any(np.allclose(trial1_spikes, trial2_spikes))


def test_clear_drives(setup_net):
    """Test clearing drives updates Network"""
    net = setup_net
    weights_ampa = {"L5_pyramidal": 0.3}
    synaptic_delays = {"L5_pyramidal": 1.0}

    # Test attributes after adding 2 drives
    n_gids = net._n_gids
    net.add_evoked_drive(
        "prox",
        mu=40,
        sigma=8.33,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=synaptic_delays,
        cell_specific=True,
    )

    net.add_evoked_drive(
        "dist",
        mu=40,
        sigma=8.33,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="distal",
        synaptic_delays=synaptic_delays,
        cell_specific=True,
    )

    for drive_name in ["prox", "dist"]:
        assert len(net.external_drives) == 2
        assert drive_name in net.external_drives
        assert drive_name in net.gid_ranges
        assert drive_name in net.pos_dict
        assert net._n_gids == n_gids + len(net.gid_ranges["L5_pyramidal"]) * 2

    # Test attributes after clearing drives
    net.clear_drives()
    for drive_name in ["prox", "dist"]:
        assert len(net.external_drives) == 0
        assert drive_name not in net.external_drives
        assert drive_name not in net.gid_ranges
        assert drive_name not in net.pos_dict
        assert net._n_gids == n_gids

    # Test attributes after adding 1 drive
    net.add_evoked_drive(
        "prox",
        mu=40,
        sigma=8.33,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="proximal",
        synaptic_delays=synaptic_delays,
        cell_specific=True,
    )

    assert len(net.external_drives) == 1
    assert "prox" in net.external_drives
    assert "prox" in net.gid_ranges
    assert "prox" in net.pos_dict
    assert net._n_gids == n_gids + len(net.gid_ranges["L5_pyramidal"])


def test_add_drives():
    """Test methods for adding drives to a Network."""
    hnn_core_root = op.dirname(hnn_core.__file__)
    params_fname = op.join(hnn_core_root, "param", "default.json")
    params = read_params(params_fname)
    net = Network(params, legacy_mode=False)

    # Ensure weights and delays are updated
    weights_ampa = {"L2_basket": 1.0, "L2_pyramidal": 3.0, "L5_pyramidal": 4.0}
    syn_delays = {"L2_basket": 1.0, "L2_pyramidal": 2.0, "L5_pyramidal": 4.0}

    n_drive_cells = 10
    cell_specific = False  # default for bursty drive
    net.add_bursty_drive(
        "bursty",
        location="distal",
        burst_rate=10,
        weights_ampa=weights_ampa,
        synaptic_delays=syn_delays,
        n_drive_cells=n_drive_cells,
    )

    assert net.external_drives["bursty"]["n_drive_cells"] == n_drive_cells
    assert net.external_drives["bursty"]["cell_specific"] == cell_specific
    conn_idxs = pick_connection(net, src_gids="bursty")
    for conn_idx in conn_idxs:
        drive_conn = net.connectivity[conn_idx]
        target_type = drive_conn["target_type"]
        assert drive_conn["nc_dict"]["A_weight"] == weights_ampa[target_type]
        assert drive_conn["nc_dict"]["A_delay"] == syn_delays[target_type]

    n_drive_cells = "n_cells"  # default for evoked drive
    cell_specific = True
    net.add_evoked_drive(
        "evoked_dist",
        mu=1.0,
        sigma=1.0,
        numspikes=1,
        weights_ampa=weights_ampa,
        location="distal",
        synaptic_delays=syn_delays,
        cell_specific=True,
    )

    n_dist_targets = 235  # 270 with legacy mode
    assert net.external_drives["evoked_dist"]["n_drive_cells"] == n_dist_targets
    assert net.external_drives["evoked_dist"]["cell_specific"] == cell_specific
    conn_idxs = pick_connection(net, src_gids="evoked_dist")
    for conn_idx in conn_idxs:
        drive_conn = net.connectivity[conn_idx]
        target_type = drive_conn["target_type"]
        assert drive_conn["nc_dict"]["A_weight"] == weights_ampa[target_type]
        assert drive_conn["nc_dict"]["A_delay"] == syn_delays[target_type]

    n_drive_cells = "n_cells"  # default for poisson drive
    cell_specific = True
    net.add_poisson_drive(
        "poisson",
        rate_constant=1.0,
        weights_ampa=weights_ampa,
        location="distal",
        synaptic_delays=syn_delays,
        cell_specific=cell_specific,
    )

    n_dist_targets = 235  # 270 with non-legacy mode
    assert net.external_drives["poisson"]["n_drive_cells"] == n_dist_targets
    assert net.external_drives["poisson"]["cell_specific"] == cell_specific
    conn_idxs = pick_connection(net, src_gids="poisson")
    for conn_idx in conn_idxs:
        drive_conn = net.connectivity[conn_idx]
        target_type = drive_conn["target_type"]
        assert drive_conn["nc_dict"]["A_weight"] == weights_ampa[target_type]
        assert drive_conn["nc_dict"]["A_delay"] == syn_delays[target_type]

    # Test drive targeting specific section
    # Section present on all cells indicated
    location = "apical_tuft"
    weights_ampa_tuft = {"L2_pyramidal": 1.0, "L5_pyramidal": 2.0}
    syn_delays_tuft = {"L2_pyramidal": 1.0, "L5_pyramidal": 2.0}
    net.add_bursty_drive(
        "bursty_tuft",
        location=location,
        burst_rate=10,
        weights_ampa=weights_ampa_tuft,
        synaptic_delays=syn_delays_tuft,
        n_drive_cells=10,
    )
    assert net.connectivity[-1]["loc"] == location

    # Section not present on cells indicated
    location = "apical_tuft"
    weights_ampa_no_tuft = {"L2_pyramidal": 1.0, "L5_basket": 2.0}
    syn_delays_no_tuft = {"L2_pyramidal": 1.0, "L5_basket": 2.0}
    match = "Invalid value for"
    with pytest.raises(ValueError, match=match):
        net.add_bursty_drive(
            "bursty_no_tuft",
            location=location,
            burst_rate=10,
            weights_ampa=weights_ampa_no_tuft,
            synaptic_delays=syn_delays_no_tuft,
            n_drive_cells=n_drive_cells,
        )

    # Test probabilistic drive connections.
    # drive with cell_specific=False
    n_drive_cells = 10
    probability = 0.5  # test that only half of possible connections are made
    weights_nmda = {"L2_basket": 1.0, "L2_pyramidal": 3.0, "L5_pyramidal": 4.0}
    net.add_bursty_drive(
        "bursty_prob",
        location="distal",
        burst_rate=10,
        weights_ampa=weights_ampa,
        weights_nmda=weights_nmda,
        synaptic_delays=syn_delays,
        n_drive_cells=n_drive_cells,
        probability=probability,
    )

    for cell_type in weights_ampa.keys():
        conn_idxs = pick_connection(net, src_gids="bursty_prob", target_gids=cell_type)
        gid_pairs_comparison = net.connectivity[conn_idxs[0]]["gid_pairs"]
        for conn_idx in conn_idxs:
            conn = net.connectivity[conn_idx]
            num_connections = np.sum([len(gids) for gids in conn["gid_pairs"].values()])
            # Ensures that AMPA and NMDA connections target the same gids.
            # Necessary when weights of both are non-zero.
            assert gid_pairs_comparison == conn["gid_pairs"]
            assert num_connections == np.around(
                len(net.gid_ranges[cell_type]) * n_drive_cells * probability
            ).astype(int)

    # drives with cell_specific=True
    probability = {"L2_basket": 0.1, "L2_pyramidal": 0.25, "L5_pyramidal": 0.5}
    net.add_evoked_drive(
        "evoked_prob",
        mu=1.0,
        sigma=1.0,
        numspikes=1,
        weights_ampa=weights_ampa,
        weights_nmda=weights_nmda,
        location="distal",
        synaptic_delays=syn_delays,
        cell_specific=True,
        probability=probability,
    )

    for cell_type in weights_ampa.keys():
        conn_idxs = pick_connection(net, src_gids="evoked_prob", target_gids=cell_type)
        gid_pairs_comparison = net.connectivity[conn_idxs[0]]["gid_pairs"]
        for conn_idx in conn_idxs:
            conn = net.connectivity[conn_idx]
            num_connections = np.sum([len(gids) for gids in conn["gid_pairs"].values()])
            assert gid_pairs_comparison == conn["gid_pairs"]
            assert num_connections == np.around(
                len(net.gid_ranges[cell_type]) * probability[cell_type]
            ).astype(int)

    # Test adding just the NMDA weights (no AMPA)
    net.add_evoked_drive(
        "evoked_nmda",
        mu=1.0,
        sigma=1.0,
        numspikes=1,
        weights_nmda=weights_nmda,
        location="distal",
        synaptic_delays=syn_delays,
        cell_specific=True,
        probability=probability,
    )

    # Round trip test to ensure drives API produces a functioning Network
    simulate_dipole(net, tstop=1)

    # evoked
    with pytest.raises(ValueError, match="Standard deviation cannot be negative"):
        net.add_evoked_drive("evdist1", mu=10, sigma=-1, numspikes=1, location="distal")
    with pytest.raises(ValueError, match="Number of spikes must be greater than zero"):
        net.add_evoked_drive("evdist1", mu=10, sigma=1, numspikes=0, location="distal")

    # Test Network._attach_drive()
    with pytest.raises(ValueError, match="Invalid value for"):
        net.add_evoked_drive(
            "evdist1",
            mu=10,
            sigma=1,
            numspikes=1,
            location="bogus_location",
            weights_ampa={"L5_basket": 1.0},
            synaptic_delays={"L5_basket": 0.1},
        )
    with pytest.raises(ValueError, match="Drive evoked_dist already defined"):
        net.add_evoked_drive(
            "evoked_dist", mu=10, sigma=1, numspikes=1, location="distal"
        )
    with pytest.raises(
        ValueError, match="No target cell types have been given a synaptic weight"
    ):
        net.add_evoked_drive("evdist1", mu=10, sigma=1, numspikes=1, location="distal")
    with pytest.raises(
        ValueError,
        match="Due to physiological/anatomical constraints, "
        "a distal drive cannot target L5_basket cell types. ",
    ):
        net.add_evoked_drive(
            "evdist1",
            mu=10,
            sigma=1,
            numspikes=1,
            location="distal",
            weights_ampa={"L5_basket": 1.0},
            synaptic_delays={"L5_basket": 0.1},
        )
    with pytest.raises(ValueError, match="If cell_specific is True, n_drive_cells"):
        net.add_evoked_drive(
            "evdist1",
            mu=10,
            sigma=1,
            numspikes=1,
            location="distal",
            n_drive_cells=10,
            cell_specific=True,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )
    with pytest.raises(ValueError, match="If cell_specific is False, n_drive_cells"):
        net.add_evoked_drive(
            "evdist1",
            mu=10,
            sigma=1,
            numspikes=1,
            location="distal",
            n_drive_cells="n_cells",
            cell_specific=False,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )
    with pytest.raises(
        ValueError, match="Number of drive cells must be greater than 0"
    ):
        net.add_evoked_drive(
            "evdist1",
            mu=10,
            sigma=1,
            numspikes=1,
            location="distal",
            n_drive_cells=0,
            cell_specific=False,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )

    # Poisson
    with pytest.raises(
        ValueError, match="End time of Poisson drive cannot be negative"
    ):
        net.add_poisson_drive(
            "poisson1", tstart=0, tstop=-1, location="distal", rate_constant=10.0
        )
    with pytest.raises(
        ValueError, match="Start time of Poisson drive cannot be negative"
    ):
        net.add_poisson_drive(
            "poisson1", tstart=-1, location="distal", rate_constant=10.0
        )
    with pytest.raises(
        ValueError, match="Duration of Poisson drive cannot be negative"
    ):
        net.add_poisson_drive(
            "poisson1", tstart=10, tstop=1, location="distal", rate_constant=10.0
        )
    with pytest.raises(ValueError, match="Rate constant must be positive"):
        net.add_poisson_drive(
            "poisson1",
            location="distal",
            rate_constant=0.0,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )

    with pytest.raises(ValueError, match="Rate constants not provided for all target"):
        net.add_poisson_drive(
            "poisson1",
            location="distal",
            rate_constant={"L2_pyramidal": 10.0},
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )
    with pytest.raises(
        ValueError, match="Rate constant provided for unknown target cell"
    ):
        net.add_poisson_drive(
            "poisson1",
            location="distal",
            rate_constant={"L2_pyramidal": 10.0, "bogus_celltype": 20.0},
            weights_ampa={"L2_pyramidal": 0.01, "bogus_celltype": 0.01},
            synaptic_delays=0.1,
        )

    with pytest.raises(
        ValueError,
        match="Drives specific to cell types are only possible with cell_specific=True",
    ):
        net.add_poisson_drive(
            "poisson1",
            location="distal",
            rate_constant={
                "L2_basket": 10.0,
                "L2_pyramidal": 11.0,
                "L5_basket": 12.0,
                "L5_pyramidal": 13.0,
            },
            n_drive_cells=1,
            cell_specific=False,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )

    # bursty
    with pytest.raises(ValueError, match="End time of bursty drive cannot be negative"):
        net.add_bursty_drive("bursty_drive", tstop=-1, location="distal", burst_rate=10)
    with pytest.raises(
        ValueError, match="Start time of bursty drive cannot be negative"
    ):
        net.add_bursty_drive(
            "bursty_drive", tstart=-1, location="distal", burst_rate=10
        )
    with pytest.raises(ValueError, match="Duration of bursty drive cannot be negative"):
        net.add_bursty_drive(
            "bursty_drive", tstart=10, tstop=1, location="distal", burst_rate=10
        )

    msg = (
        r"(?s)Burst duration .* cannot be greater than "
        "burst period"
    )
    with pytest.raises(ValueError, match=msg):
        net.add_bursty_drive(
            "bursty_drive",
            location="distal",
            burst_rate=10,
            burst_std=20.0,
            numspikes=4,
            spike_isi=50,
        )

    # attaching drives
    with pytest.raises(ValueError, match="Drive evoked_dist already defined"):
        net.add_poisson_drive(
            "evoked_dist",
            location="distal",
            rate_constant=10.0,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )
    with pytest.raises(ValueError, match="Invalid value for the"):
        net.add_poisson_drive(
            "weird_poisson",
            location="between",
            rate_constant=10.0,
            weights_ampa=weights_ampa,
            synaptic_delays=syn_delays,
        )
    with pytest.raises(ValueError, match="Allowed drive target cell types are:"):
        net.add_poisson_drive(
            "cell_unknown",
            location="proximal",
            rate_constant=10.0,
            weights_ampa={"CA1_pyramidal": 1.0},
            synaptic_delays=0.01,
        )
    with pytest.raises(
        ValueError,
        match="synaptic_delays is either a common float or "
        "needs to be specified as a dict for each of the cell",
    ):
        net.add_poisson_drive(
            "cell_unknown",
            location="proximal",
            rate_constant=10.0,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L5_pyramidal": 1.0},
        )
    with pytest.raises(ValueError, match=r"probability must be in the range \(0\,1\)"):
        net.add_bursty_drive(
            "cell_unknown",
            location="distal",
            burst_rate=10,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L2_pyramidal": 1.0},
            probability=2.0,
        )

    with pytest.raises(
        TypeError,
        match="probability must be an instance of "
        r"float or dict, got \<class 'str'\> instead",
    ):
        net.add_bursty_drive(
            "cell_unknown2",
            location="distal",
            burst_rate=10,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L2_pyramidal": 1.0},
            probability="1.0",
        )

    with pytest.raises(
        ValueError,
        match="probability is either a common "
        "float or needs to be specified as a dict for "
        "each of the cell",
    ):
        net.add_bursty_drive(
            "cell_unknown2",
            location="distal",
            burst_rate=10,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L2_pyramidal": 1.0},
            probability={"L5_pyramidal": 1.0},
        )

    with pytest.raises(
        TypeError,
        match="probability must be an instance of "
        r"float, got \<class 'str'\> instead",
    ):
        net.add_bursty_drive(
            "cell_unknown2",
            location="distal",
            burst_rate=10,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L2_pyramidal": 1.0},
            probability={"L2_pyramidal": "1.0"},
        )

    with pytest.raises(ValueError, match=r"probability must be in the range \(0\,1\)"):
        net.add_bursty_drive(
            "cell_unknown3",
            location="distal",
            burst_rate=10,
            weights_ampa={"L2_pyramidal": 1.0},
            synaptic_delays={"L2_pyramidal": 1.0},
            probability={"L2_pyramidal": 2.0},
        )

    with pytest.warns(UserWarning, match="No external drives or biases load"):
        net.clear_drives()
        simulate_dipole(net, tstop=10)


def test_drive_random_state():
    """Tests to check same random state always gives same spike times."""

    weights_ampa = {
        "L2_basket": 0.08,
        "L2_pyramidal": 0.02,
        "L5_basket": 0.2,
        "L5_pyramidal": 0.00865,
    }
    synaptic_delays = {
        "L2_basket": 0.1,
        "L2_pyramidal": 0.1,
        "L5_basket": 1.0,
        "L5_pyramidal": 1.0,
    }

    net = jones_2009_model()
    for drive_name in ["evprox1", "evprox2"]:
        net.add_evoked_drive(
            drive_name,
            mu=137.12,
            sigma=8,
            numspikes=1,
            weights_ampa=weights_ampa,
            weights_nmda=None,
            location="proximal",
            synaptic_delays=synaptic_delays,
            event_seed=4,
        )

    net._instantiate_drives(tstop=170.0)
    assert (
        net.external_drives["evprox1"]["events"]
        == net.external_drives["evprox2"]["events"]
    )


@pytest.mark.parametrize(
    "rate_constant,cell_specific,n_drive_cells",
    [
        (2, False, 1),
        (2.0, False, 1),
        (2, True, "n_cells"),
        (2.0, True, "n_cells"),
    ],
)
def test_add_poisson_drive(setup_net, rate_constant, cell_specific, n_drive_cells):
    """Testing rate constant when adding non-cell-specific poisson drive"""
    net = setup_net

    weights_ampa_noise = {
        "L2_basket": 0.01,
        "L2_pyramidal": 0.002,
        "L5_pyramidal": 0.02,
    }

    net.add_poisson_drive(
        "noise_global",
        rate_constant=rate_constant,
        location="distal",
        weights_ampa=weights_ampa_noise,
        space_constant=100,
        n_drive_cells=n_drive_cells,
        cell_specific=cell_specific,
    )

    simulate_dipole(net, tstop=5)



"""Establish class def for general cell features."""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>

from copy import deepcopy

import numpy as np
from numpy.linalg import norm

from neuron import h, nrn

from .viz import plot_cell_morphology
from .externals.mne import _validate_type, _check_option

# Units for e: mV
# Units for gbar: S/cm^2


def _get_cos_theta(sections, sec_name_apical):
    """Get cos(theta) to compute dipole along the apical dendrite."""
    a = np.array(sections[sec_name_apical].end_pts[1]) - np.array(
        sections[sec_name_apical].end_pts[0]
    )
    cos_thetas = dict()
    for sec_name, section in sections.items():
        b = np.array(section.end_pts[1]) - np.array(section.end_pts[0])
        cos_thetas[sec_name] = np.dot(a, b) / (norm(a) * norm(b))
    return cos_thetas


def _calculate_gaussian(x_val, height, lamtha):
    """Return height of gaussian at x_val.

    Parameters
    ----------
    x_val : float
        Value on x-axis to query height of gaussian curve.
    height : float
        Height of the gaussian curve at zero.
    lamtha : float
        Space constant.

    Returns
    -------
    x_height : float
        Height of gaussian at x_val.

    Notes
    -----
    Gaussian curve is centered at zero and has a fixed peak height
    such the _calculate_gaussian(0, lamtha) returns 1 for all lamtha.
    """
    x_height = height * np.exp(-(x_val**2) / (lamtha**2))

    return x_height


def _get_gaussian_connection(src_pos, target_pos, nc_dict, inplane_distance=1.0):
    """Calculate distance dependent connection properties.

    Parameters
    ----------
    src_pos : float
        Position of source cell.
    target_pos : float
        Position of target cell.
    nc_dict : dict
        Dictionary with keys: pos_src, A_weight, A_delay, lamtha
        Defines the connection parameters
    inplane_distance : float
        The in plane-distance (in um) between pyramidal cell somas in the
        square grid. Default: 1.0 um.

    Returns
    -------
    weight : float
        Weight of the synaptic connection.
    delay : float
        Delay of synaptic connection.

    Notes
    -----
    Distance in xy plane is used for gaussian decay.
    """
    x_dist = target_pos[0] - src_pos[0]
    y_dist = target_pos[1] - src_pos[1]
    cell_dist = np.sqrt(x_dist**2 + y_dist**2)
    scaled_lamtha = nc_dict["lamtha"] * inplane_distance

    weight = _calculate_gaussian(cell_dist, nc_dict["A_weight"], scaled_lamtha)
    delay = nc_dict["A_delay"] / _calculate_gaussian(cell_dist, 1, scaled_lamtha)
    return weight, delay


def node_to_str(node):
    return node[0] + "," + str(node[1])


class _ArtificialCell:
    """The ArtificialCell class for initializing a NEURON feed source.

    Parameters
    ----------
    event_times : list
        Spike times associated with a single feed source (i.e.,
        associated with a unique gid).
    threshold : float
        Membrane potential threshold that demarks a spike.
    gid : int or None (optional)
        Each cell in a network is uniquely identified by it's "global ID": GID.
        The GID is an integer from 0 to n_cells, or None if the cell is not
        yet attached to a network. Once the GID is set, it cannot be changed.

    Attributes
    ----------
    nrn_eventvec : instance of h.Vector()
        NEURON h.Vector() object of event times.
    nrn_vecstim : instance of h.VecStim()
        NEURON h.VecStim() object of spike sources created
        from nrn_eventvec.
    nrn_netcon : instance of h.NetCon()
        NEURON h.NetCon() object that creates the spike
        source-to-target references for nrn_vecstim.
    gid : int
        GID of the cell in a network (or None if not yet assigned)
    """

    def __init__(self, event_times, threshold, gid=None):
        # Convert event times into nrn vector
        self.nrn_eventvec = h.Vector()
        self.nrn_eventvec.from_python(event_times)

        # load eventvec into VecStim object
        self.nrn_vecstim = h.VecStim()
        self.nrn_vecstim.play(self.nrn_eventvec)

        # create the cell and artificial NetCon
        self.nrn_netcon = h.NetCon(self.nrn_vecstim, None)
        self.nrn_netcon.threshold = threshold

        self._gid = None
        if gid is not None:
            self.gid = gid  # use setter method to check input argument gid

    @property
    def gid(self):
        return self._gid

    @gid.setter
    def gid(self, gid):
        if not isinstance(gid, int):
            raise ValueError("gid must be an integer")
        if self._gid is None:
            self._gid = gid
        else:
            raise RuntimeError("Global ID for this cell already assigned!")


def _get_nseg(L):
    nseg = 1
    if L > 100.0:  # 100 um
        nseg = int(L / 50.0)
        # make dend.nseg odd for all sections
        if not nseg % 2:
            nseg += 1
    return nseg


class Section:
    """Section class.

    Parameters
    ----------
    L : float
        length of a section in microns.
    diam : float
        diameter of a section in microns.
    cm : float
        membrane capacitance in micro-Farads.
    Ra : float
        axial resistivity in ohm-cm
    end_pts : list of [x, y, z]
        The start and stop points of the section.

    Attributes
    ----------
    mechs : dict
        Mechanisms to insert in this section. The keys
        are the names of the mechanisms and values
        are the properties. For e.g., {'ca': {'gbar_ca': 60}}
    syns : list of str
        The synaptic mechanisms to add in this section
    end_pts : list of [x, y, z]
        The start and stop points of the section. Cannot be changed.
    L : float
        length of a section in microns.
    diam : float
        diameter of a section in microns.
    cm : float
        membrane capacitance in micro-Farads.
    Ra : float
        axial resistivity in ohm-cm.
    nseg : int
        Number of segments in the section
    """

    def __init__(self, L, diam, Ra, cm, end_pts=None):
        self._L = L
        self._diam = diam
        self._Ra = Ra
        self._cm = cm
        if end_pts is None:
            end_pts = list()
        self._end_pts = end_pts

        self.mechs = dict()
        self.syns = list()

        # For distance functionality
        self.nseg = _get_nseg(self.L)

    def __repr__(self):
        return f"L={self.L}, diam={self.diam}, cm={self.cm}, Ra={self.Ra}"

    def __eq__(self, other):
        if not isinstance(other, Section):
            return NotImplemented

        # Check equality for mechs
        for mech_name in self.mechs.keys():
            self_mech = self.mechs[mech_name]
            other_mech = other.mechs[mech_name]
            for attr in self_mech.keys():
                if self_mech[attr] != other_mech[attr]:
                    return False

        # Check end_pts
        for self_end_pt, other_end_pt in zip(self.end_pts, other.end_pts):
            if np.testing.assert_almost_equal(self_end_pt, other_end_pt, 5) is not None:
                return False

        all_attrs = dir(self)
        attrs_to_ignore = [x for x in all_attrs if x.startswith("_")]
        attrs_to_ignore.extend(["end_pts", "mechs", "to_dict"])
        attrs_to_check = [x for x in all_attrs if x not in attrs_to_ignore]

        # Check all other attributes
        for attr in attrs_to_check:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def to_dict(self):
        """Converts an object of Section class to a dictionary.

        Returns
        -------
        dictionary form of an object of Section class.
        """
        section_data = dict()
        section_data["L"] = self.L
        section_data["diam"] = self.diam
        section_data["cm"] = self.cm
        section_data["Ra"] = self.Ra
        section_data["end_pts"] = self.end_pts
        section_data["nseg"] = self.nseg
        # Need to solve the partial function problem
        # in mechs
        section_data["mechs"] = self.mechs
        section_data["syns"] = self.syns
        return section_data

    @property
    def L(self):
        return self._L

    @property
    def diam(self):
        return self._diam

    @property
    def cm(self):
        return self._cm

    @property
    def Ra(self):
        return self._Ra

    @property
    def end_pts(self):
        return self._end_pts


class Cell:
    """Create a cell object.

    Parameters
    ----------
    name : str
        The name of the cell.
    pos : tuple
        The (x, y, z) coordinates.
    sections : dict of Section
        Dictionary with keys as section name.
    synapses : dict of dict
        Keys are name of synaptic mechanism. Each synaptic mechanism
        has keys for parameters of the mechanism, e.g., 'e', 'tau1',
        'tau2'.
        sections.
    sect_loc : dict of list
        Can have keys 'proximal' or 'distal' each containing
        names of section locations that are proximal or distal.
    gid : int or None (optional)
        Each cell in a network is uniquely identified by it's "global ID": GID.
        The GID is an integer from 0 to n_cells, or None if the cell is not
        yet attached to a network. Once the GID is set, it cannot be changed.
    cell_tree : dict of list
        Stores the tree representation of a cell.
        Root is the 0 end of 'soma'. Nodes are a tuple (sec_name, node_pos)
        where sec_name is the name of the section and node_pos is the 0 end
        or 1 end. The data structure is the adjacency list representation of a
        tree. The keys of the dict are the parent nodes. The value is the
        list of nodes (children nodes) connected to the parent node.


    Attributes
    ----------
    pos : list of length 3
        The position of the cell.
    sections : nested dict
        The section parameters. The key is the name of the section
        and the value is a dictionary parametrizing the morphology
        of the section and the mechanisms inserted.
    synapses : dict
        The synapses that the cell can use for connections.
    dipole_pp : list of h.Dipole()
        The Dipole objects (see dipole.mod).
    vsec : dict
        Recording of section specific voltage. Must be enabled
        by running simulate_dipole(net, record_vsec=True) or
        simulate_dipole(net, record_vsoma=True)
    isec : dict
        Contains recording of section specific currents indexed
        by synapse type (keys can be soma_gabaa, soma_gabab etc.).
        Must be enabled by running simulate_dipole(net, record_isec=True)
        or simulate_dipole(net, record_isoma=True)
    ca : dict
        Contains recording of section speicifc calcium concentration.
        Must be enabled by running simulate_dipole(net, record_ca=True).
    tonic_biases : list of h.IClamp
        The current clamps inserted at each section of the cell
        for tonic biasing inputs.
    gid : int
        GID of the cell in a network (or None if not yet assigned)
    sect_loc : dict of list
        Can have keys 'proximal' or 'distal' each containing
        names of section locations that are proximal or distal.
    cell_tree : dict of list
        Stores the tree representation of a cell.
        Root is the 0 end of 'soma'. Nodes are a tuple (sec_name, node_pos)
        where sec_name is the name of the section and node_pos is the 0 end
        or 1 end. The data structure is the adjacency list representation of a
        tree. The keys of the dict are the parent nodes. The value is the
        list of nodes (children nodes) connected to the parent node.

    Examples
    --------
    >>> section_soma = Section(
            L=39,
            diam=20,
            cm=0.85,
            Ra=200.,
            end_pts=[[0, 0, 0], [0, 39., 0]]
        )
    """

    def __init__(self, name, pos, sections, synapses, sect_loc, cell_tree, gid=None):
        self.name = name
        self.pos = pos
        for section in sections.values():
            if not isinstance(section, Section):
                raise ValueError(
                    f"Items in section must be instances"
                    f" of Section. Got {type(section)}"
                )
        self.sections = sections
        self.synapses = synapses
        self.sect_loc = sect_loc
        self._nrn_sections = dict()
        self._nrn_synapses = dict()
        self.dipole_pp = list()
        self.vsec = dict()
        self.isec = dict()
        self.ca = dict()
        # insert iclamp
        self.list_IClamp = list()
        self._gid = None
        self.tonic_biases = list()
        if gid is not None:
            self.gid = gid  # use setter method to check input argument gid

        # Store the tree representation of the cell
        self.cell_tree = cell_tree

        self._update_end_pts()  # New implementation

        self._compute_section_mechs()  # Set mech values of all sections

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"<{class_name} | gid={self._gid}>"

    def __eq__(self, other):
        if not isinstance(other, Cell):
            return NotImplemented

        all_attrs = dir(self)
        attrs_to_ignore = [x for x in all_attrs if x.startswith("_")]
        attrs_to_ignore.extend(
            [
                "build",
                "copy",
                "create_tonic_bias",
                "define_shape",
                "distance_section",
                "gid",
                "list_IClamp",
                "modify_section",
                "parconnect_from_src",
                "plot_morphology",
                "record",
                "sections",
                "setup_source_netcon",
                "syn_create",
                "to_dict",
            ]
        )
        attrs_to_check = [x for x in all_attrs if x not in attrs_to_ignore]

        # Check all other attributes
        for attr in attrs_to_check:
            if getattr(self, attr) != getattr(other, attr):
                return False

        if not (self.sections.keys() == other.sections.keys()):
            return False

        for key in self.sections.keys():
            if self.sections[key] != other.sections[key]:
                return False

        return True

    def to_dict(self):
        """Converts an object of Cell class to a dictionary.

        Returns
        -------
        dictionary form of an object of Cell class.
        """
        cell_data = dict()
        cell_data["name"] = self.name
        cell_data["pos"] = self.pos
        cell_data["sections"] = dict()
        for key in self.sections:
            cell_data["sections"][key] = self.sections[key].to_dict()
        cell_data["synapses"] = self.synapses
        # cell_data['cell_tree'] = self.cell_tree
        if self.cell_tree is None:
            cell_data["cell_tree"] = None
        else:
            cell_tree_dict = dict()
            for parent, children in self.cell_tree.items():
                key = node_to_str(parent)
                value = list()
                for child in children:
                    value.append(node_to_str(child))
                cell_tree_dict[key] = value
            cell_data["cell_tree"] = cell_tree_dict
        cell_data["sect_loc"] = self.sect_loc
        cell_data["gid"] = self.gid
        cell_data["dipole_pp"] = self.dipole_pp
        cell_data["vsec"] = self.vsec
        cell_data["isec"] = self.isec
        cell_data["ca"] = self.ca
        cell_data["tonic_biases"] = self.tonic_biases
        return cell_data

    @property
    def gid(self):
        return self._gid

    @gid.setter
    def gid(self, gid):
        if not isinstance(gid, int):
            raise ValueError("gid must be an integer")
        if self._gid is None:
            self._gid = gid
        else:
            raise RuntimeError("Global ID for this cell already assigned!")

    def distance_section(self, target_sec_name, curr_node):
        """Find distance between the current node and the target section.

        Parameters
        ----------
        target_sec_name : string
            Name of the target section
        curr_node : tuple
            Source node from where search begins.
            It is of the the form (sec_name, end_pt).

        Returns
        -------
        distance : float
            Path distance between source node and mid of the target section.
        """
        # Python version of the Neuron distance function
        # https://nrn.readthedocs.io/en/latest/python/modelspec/programmatic/topology/geometry.html#distance  # noqa
        if self.cell_tree is None:
            raise TypeError("distance_section() cannot work with cell_tree as None.")
        if curr_node not in self.cell_tree:
            return np.nan

        # Children of the current section
        curr_sec_children = self.cell_tree[curr_node]
        # All sections have 0 and 1 ends
        end_pts = (0, 1)

        # Base condition
        # If target section is connected to current section
        # Return (target section length / 2)
        # As distances are measured till the centre of the target section
        for end_pt in end_pts:
            if (target_sec_name, end_pt) in curr_sec_children:
                return self.sections[target_sec_name].L / 2

        dist = np.nan  # Return nan

        # Recursion to find distance
        for node in self.cell_tree[curr_node]:
            if node[0] == curr_node[0]:
                dist_temp = (
                    self.distance_section(target_sec_name, node)
                    + self.sections[node[0]].L
                )
            else:
                dist_temp = self.distance_section(target_sec_name, node)
            if np.isnan(dist) and np.isnan(dist_temp):
                dist = np.nan
            else:
                dist = np.nanmin([dist, dist_temp])

        return dist

    def _set_biophysics(self, sections):
        "Set the biophysics for the cell."

        # neuron syntax is used to set values for mechanisms
        # sec.gbar_mech = x sets value of gbar for mech to x for all segs
        # in a section. This method is significantly faster than using
        # a for loop to iterate over all segments to set mech values

        # If value depends on distance from the soma. Soma is set as
        # origin by passing cell.soma as a sec argument to h.distance()
        # Then iterate over segment nodes of dendritic sections
        # and set attribute depending on h.distance(seg.x), which returns
        # distance from the soma to this point on the CURRENTLY ACCESSED
        # SECTION!!!
        h.distance(sec=self._nrn_sections["soma"])
        for sec_name, section in sections.items():
            sec = self._nrn_sections[sec_name]
            for mech_name, p_mech in section.mechs.items():
                sec.insert(mech_name)
                for attr, val in p_mech.items():
                    if isinstance(val, list):
                        seg_xs, seg_vals = val[0], val[1]
                        for seg, seg_x, seg_val in zip(sec, seg_xs, seg_vals):
                            setattr(seg, attr, seg_val)
                    else:
                        setattr(sec, attr, val)

    def _compute_section_mechs(self):
        sections = self.sections
        for sec_name, section in sections.items():
            for mech_name, p_mech in section.mechs.items():
                for attr, val in p_mech.items():
                    if hasattr(val, "__call__"):
                        seg_xs, seg_vals = list(), list()
                        section_distance = self.distance_section(sec_name, ("soma", 0))
                        seg_centers = np.linspace(0, 1, section.nseg * 2 + 1)[1::2]

                        for seg_x in seg_centers:
                            # sec_end_dist is distance between 0 end of soma to
                            # the 0 or 1 end of section (whichever is closer)
                            sec_end_dist = section_distance - (section.L / 2)
                            seg_xs.append(seg_x)
                            seg_vals.append(val(sec_end_dist + (seg_x * section.L)))
                        p_mech[attr] = [seg_xs, seg_vals]
        return self.sections

    def _create_synapses(self, sections, synapses):
        """Create synapses."""
        for sec_name in sections:
            for receptor in sections[sec_name].syns:
                syn_key = f"{sec_name}_{receptor}"
                seg = self._nrn_sections[sec_name](0.5)
                self._nrn_synapses[syn_key] = self.syn_create(seg, **synapses[receptor])

    def _create_sections(self, sections, cell_tree):
        """Create soma and set geometry.

        Notes
        -----
        By default neuron uses xy plane
        for height and xz plane for depth. This is opposite for model as a
        whole, but convention is followed in this function ease use of gui.
        """
        if "soma" not in self.sections:
            raise KeyError("soma must be defined for cell")

        for sec_name in sections:
            sec = h.Section(name=f"{self.name}_{sec_name}")
            self._nrn_sections[sec_name] = sec

            h.pt3dclear(sec=sec)
            h.pt3dconst(0, sec=sec)  # be explicit, see documentation
            for pt in sections[sec_name].end_pts:
                # Add positional offset to each coordinate
                h.pt3dadd(
                    pt[0] + self.pos[0],
                    pt[1] + self.pos[1],
                    pt[2] + self.pos[2],
                    1,
                    sec=sec,
                )
            # with pt3dconst==0, these will alter the 3d points defined above!
            sec.L = sections[sec_name].L
            sec.diam = sections[sec_name].diam
            sec.Ra = sections[sec_name].Ra
            sec.cm = sections[sec_name].cm
            sec.nseg = sections[sec_name].nseg

        if cell_tree is None:
            cell_tree = dict()

        # Connects sections of THIS cell together.
        for parent_node in cell_tree:
            for child_node in cell_tree[parent_node]:
                parent_sec = self._nrn_sections[parent_node[0]]
                child_sec = self._nrn_sections[child_node[0]]
                if parent_sec == child_sec:
                    continue
                parent_loc = parent_node[1]
                child_loc = child_node[1]
                child_sec.connect(parent_sec, parent_loc, child_loc)

        # be explicit about letting sec.L dominate over the 3d points used by
        # h.pt3dadd(); see
        # https://nrn.readthedocs.io/en/latest/python/modelspec/programmatic/topology/geometry.html?highlight=pt3dadd#pt3dadd  # noqa
        h.define_shape()

    def build(self, sec_name_apical=None):
        """Build cell in Neuron and insert dipole if applicable.

        Parameters
        ----------
        sec_name_apical : str | None
            If not None, a dipole will be inserted in this cell in alignment
            with this section. The section should belong to the apical dendrite
            of a pyramidal neuron.
        """
        self._create_sections(self.sections, self.cell_tree)
        self._create_synapses(self.sections, self.synapses)
        self._set_biophysics(self.sections)
        if sec_name_apical in self._nrn_sections:
            self._insert_dipole(sec_name_apical)
        elif sec_name_apical is not None:
            raise ValueError(
                f"sec_name_apical must be an existing "
                f"section of the current cell or None. "
                f"Got {sec_name_apical}."
            )

    def copy(self):
        """Return copy of instance."""
        return deepcopy(self)

    # two things need to happen here for h:
    # 1. dipole needs to be inserted into each section
    # 2. a list needs to be created with a Dipole (Point Process) in each
    #    section at position 1
    # In Cell() and not Pyr() for future possibilities
    def _insert_dipole(self, sec_name_apical):
        """Insert dipole into each section of this cell.

        Parameters
        ----------
        sec_name_apical : str
            The name of the section along which dipole moment is calculated.
        """
        self.dpl_vec = h.Vector(1)
        self.dpl_ref = self.dpl_vec._ref_x[0]
        cos_thetas = _get_cos_theta(self.sections, "apical_trunk")

        # setting pointers and ztan values
        for sect_name in self.sections:
            sect = self._nrn_sections[sect_name]
            sect.insert("dipole")

            dpp = h.Dipole(1, sec=sect)  # defined in dipole_pp.mod
            self.dipole_pp.append(dpp)
            dpp.ri = h.ri(1, sec=sect)  # assign internal resistance
            # sets pointers in dipole mod file to the correct locations
            dpp._ref_pv = sect(0.99)._ref_v
            dpp._ref_Qtotal = self.dpl_ref
            # gives INTERNAL segments of the section, non-endpoints
            # creating this because need multiple values simultaneously
            pos_all = np.array([seg.x for seg in sect.allseg()])
            seg_lens = np.diff(pos_all) * sect.L
            seg_lens_z = seg_lens * cos_thetas[sect_name]

            # alternative procedure below with y_long(itudinal)
            # y_long = (h.y3d(1, sec=sect) - h.y3d(0, sec=sect)) * pos
            # y_diff = np.diff(y_long)

            # doing range to index multiple values of the same
            # np.array simultaneously
            for idx, pos in enumerate(pos_all[1:-1]):
                # assign the ri value to the dipole
                # ri not defined at 0 and L
                sect(pos).dipole.ri = h.ri(pos, sec=sect)
                # range variable 'dipole'
                # set pointers to previous segment's voltage, with
                # boundary condition
                sect(pos).dipole._ref_pv = sect(pos_all[idx])._ref_v

                # set aggregate pointers
                sect(pos).dipole._ref_Qsum = dpp._ref_Qsum
                sect(pos).dipole._ref_Qtotal = self.dpl_ref
                # add ztan values
                sect(pos).dipole.ztan = seg_lens_z[idx]
            # set the pp dipole's ztan value to the last value from seg_lens_z
            dpp.ztan = seg_lens_z[-1]
        self.dipole = h.Vector().record(self.dpl_ref)

    def create_tonic_bias(self, amplitude, t0, tstop, section="soma", loc=0.5):
        """Create tonic bias at defined section.

        Parameters
        ----------
        amplitude : float
            The amplitude of the input.
        t0 : float
            The start time of tonic input (in ms).
        tstop : float
            The end time of tonic input (in ms).
        section : str
            Section tonic input is applied to.
        loc : float (0 to 1)
            The location of the input in the section.
        """

        stim = h.IClamp(self._nrn_sections[section](loc))
        stim.delay = t0
        stim.dur = tstop - t0
        stim.amp = amplitude
        self.tonic_biases.append(stim)

    def record(self, record_vsec=False, record_isec=False, record_ca=False):
        """Record current and voltage from all sections

        Parameters
        ----------
        record_vsec : 'all' | 'soma' | False
            Option to record voltages from all sections ('all'), or just
            the soma ('soma'). Default: False.
        record_isec : 'all' | 'soma' | False
            Option to record voltages from all sections ('all'), or just
            the soma ('soma'). Default: False.
        record_ca : 'all' | 'soma' | False
            Option to record calcium concentration from all sections ('all'),
            or just the soma ('soma'). Default: False.
        """

        section_names = list(self.sections.keys())

        # Logic checks if just recording soma, sections, or both
        if record_vsec == "soma":
            self.vsec = dict.fromkeys(["soma"])
        elif record_vsec == "all":
            self.vsec = dict.fromkeys(section_names)

        if record_vsec:
            for sec_name in self.vsec:
                self.vsec[sec_name] = h.Vector()
                self.vsec[sec_name].record(self._nrn_sections[sec_name](0.5)._ref_v)

        if record_isec == "soma":
            self.isec = dict.fromkeys(["soma"])
        elif record_isec == "all":
            self.isec = dict.fromkeys(section_names)

        if record_isec:
            for sec_name in self.isec:
                list_syn = [
                    key
                    for key in self._nrn_synapses.keys()
                    if key.startswith(f"{sec_name}_")
                ]
                self.isec[sec_name] = dict.fromkeys(list_syn)

                for syn_name in self.isec[sec_name]:
                    self.isec[sec_name][syn_name] = h.Vector()
                    self.isec[sec_name][syn_name].record(
                        self._nrn_synapses[syn_name]._ref_i
                    )

        # calcium concentration
        if record_ca == "soma":
            self.ca = dict.fromkeys(["soma"])
        elif record_ca == "all":
            self.ca = dict.fromkeys(section_names)

        if record_ca:
            for sec_name in self.ca:
                if hasattr(self._nrn_sections[sec_name](0.5), "_ref_cai"):
                    self.ca[sec_name] = h.Vector()
                    self.ca[sec_name].record(self._nrn_sections[sec_name](0.5)._ref_cai)

    def syn_create(self, secloc, e, tau1, tau2):
        """Create an h.Exp2Syn synapse.

        Parameters
        ----------
        secloc : instance of nrn.Segment
            The section location. E.g., soma(0.5).
        e: float
            Reverse potential (in mV)
        tau1: float
            Rise time (in ms)
        tau2: float
            Decay time (in ms)

        Returns
        -------
        syn : instance of h.Exp2Syn
            A two state kinetic scheme synapse.
        """
        if not isinstance(secloc, nrn.Segment):
            raise TypeError(
                f"secloc must be instance ofnrn.Segment. Got {type(secloc)}"
            )
        syn = h.Exp2Syn(secloc)
        syn.e = e
        syn.tau1 = tau1
        syn.tau2 = tau2
        return syn

    def setup_source_netcon(self, threshold):
        """Created for _PC.cell and specifies SOURCES of spikes.

        Parameters
        ----------
        threshold : float
            The voltage threshold for action potential.
        """
        nc = h.NetCon(
            self._nrn_sections["soma"](0.5)._ref_v, None, sec=self._nrn_sections["soma"]
        )
        nc.threshold = threshold
        return nc

    def parconnect_from_src(self, gid_presyn, nc_dict, postsyn, inplane_distance):
        """Parallel receptor-centric connect FROM presyn TO this cell,
           based on GID.

        Parameters
        ----------
        gid_presyn : int
            The cell ID of the presynaptic neuron
        nc_dict : dict
            Dictionary with keys: pos_src, A_weight, A_delay, lamtha
            Defines the connection parameters
        postsyn : instance of h.Exp2Syn
            The postsynaptic cell object.
        inplane_distance : float
            The in plane-distance (in um) between pyramidal cell somas in the
            square grid.

        Returns
        -------
        nc : instance of h.NetCon
            A network connection object.
        """
        from .network_builder import _PC

        nc = _PC.gid_connect(gid_presyn, postsyn)

        # set props here.
        nc.threshold = nc_dict["threshold"]
        nc.weight[0], nc.delay = _get_gaussian_connection(
            nc_dict["pos_src"], self.pos, nc_dict, inplane_distance=inplane_distance
        )

        return nc

    def plot_morphology(
        self,
        ax=None,
        color=None,
        pos=(0, 0, 0),
        xlim=(-250, 150),
        ylim=(-100, 100),
        zlim=(-100, 1200),
        show=True,
    ):
        """Plot the cell morphology.

        Parameters
        ----------
        ax : instance of Axes3D
            Matplotlib 3D axis
        color : str | dict | None
            Color of cell. If str, entire cell plotted with
            color indicated by str. If dict, colors of individual sections
            can be specified. Must have a key for every section in cell as
            defined in the `Cell.sections` attribute.

        | Ex: ``{'apical_trunk': 'r', 'soma': 'b', ...}``
        pos : tuple of int or float | None
            Position of cell soma. Must be a tuple of 3 elements for the
            (x, y, z) position of the soma in 3D space. Default: (0, 0, 0)
        xlim : tuple of int | tuple of float
            x limits of plot window. Default (-250, 150)
        ylim : tuple of int | tuple of float
            y limits of plot window. Default (-100, 100)
        zlim : tuple of int | tuple of float
            z limits of plot window. Default (-100, 1200)
        show : bool
            If True, show the plot

        Returns
        -------
        axes : instance of Axes3D
            The matplotlib 3D axis handle.
        """
        return plot_cell_morphology(
            self,
            ax=ax,
            color=color,
            pos=pos,
            xlim=xlim,
            ylim=ylim,
            zlim=zlim,
            show=show,
        )

    def _update_section_end_pts_L(self, node, dpt):
        if self.cell_tree is None:
            return
        x = self.sections[node[0]].end_pts[node[1]][0]
        y = self.sections[node[0]].end_pts[node[1]][1]
        z = self.sections[node[0]].end_pts[node[1]][2]
        self.sections[node[0]].end_pts[node[1]][0] = x + dpt[0]
        self.sections[node[0]].end_pts[node[1]][1] = y + dpt[1]
        self.sections[node[0]].end_pts[node[1]][2] = z + dpt[2]

        # If current node is a leaf node
        if node not in self.cell_tree:
            return

        # If current node is an internal node
        for child_node in self.cell_tree[node]:
            self._update_section_end_pts_L(child_node, dpt)

    def define_shape(self, node):
        """Redefines end_pts according to section lengths.

        Detects change in section lengths of the sections in the
        subtree of the input node.

        Parameters
        ----------
        node : tuple of size 2
            The first element is the section name
            The second element is the node end used (0 or 1)

        Note
        ----
        Using sec_name as 'soma' and node end as 0 checks for changes
        in any section length of the cell as (soma, 0) is the root node
        of the cell.
        """
        # Python version of Neuron define_shape function
        # https://nrn.readthedocs.io/en/latest/python/modelspec/programmatic/topology/geometry.html?highlight=pt3dadd#pt3dadd  # noqa
        # cell tree is None therefore cannot define shape
        if self.cell_tree is None:
            return
        # Find the end pts of the section
        node_opp_end = 1
        if node[1] == 1:
            node_opp_end = 0
        pts = self.sections[node[0]].end_pts
        x0, y0, z0 = pts[node[1]][0], pts[node[1]][1], pts[node[1]][2]
        x1, y1, z1 = (pts[node_opp_end][0], pts[node_opp_end][1], pts[node_opp_end][2])

        # Find the factor by which length is changed
        end_1 = np.array((x0, y0, z0))
        end_2 = np.array((x1, y1, z1))
        old_len = np.linalg.norm(end_1 - end_2)
        new_len = self.sections[node[0]].L
        fac = new_len / old_len
        x_new = x0 + (x1 - x0) * fac
        y_new = y0 + (y1 - y0) * fac
        z_new = z0 + (z1 - z0) * fac

        # Find the change in coordinates
        dx = x_new - x1
        dy = y_new - y1
        dz = z_new - z1
        dpt = [dx, dy, dz]

        # Update all coordinates in the subtree
        self._update_section_end_pts_L((node[0], node_opp_end), dpt)

        # Check for change in section lengths in the subtree
        if node in self.cell_tree:
            for child_node in self.cell_tree[node]:
                self.define_shape(child_node)

    def _update_end_pts(self):
        """Update all end pts according to the length of the sections.

        Can be used whenever length of any section is updated

        Returns
        -------
        Updated end pts for the cell
        """
        if "soma" not in self.sections:
            raise KeyError("soma must be defined for cell")
        # cell tree is None therefore no end_pts to update
        if self.cell_tree is None:
            return

        # shift cell to self.pos and reorient apical dendrite
        # along z direction of self.pos
        dx = self.pos[0] - self.sections["soma"].end_pts[0][0]
        dy = self.pos[1] - self.sections["soma"].end_pts[0][1]
        dz = self.pos[2] - self.sections["soma"].end_pts[0][2]
        for sec_name in self.sections:
            end_pts = self.sections[sec_name].end_pts
            updated_end_pts = list()
            for pt in end_pts:
                updated_end_pts.append([pt[0] + dx, pt[1] + dy, pt[2] + dz])
            self.sections[sec_name]._end_pts = updated_end_pts

        # Check and update all end pts starting from root according to length
        # of sections.
        self.define_shape(("soma", 0))

    def modify_section(self, sec_name, L=None, diam=None, cm=None, Ra=None):
        """Change attributes of section specified by `sec_name`

        Parameters
        ----------
        sec_name : str
            Name of section to be modified. Must be a key of Cell.sections
        L : float | int | None
            length of a section in microns. Default None.
        diam : float | int | None
            diameter of a section in microns.
        cm : float | int | None
            membrane capacitance in micro-Farads.
        Ra : float | int | None
            axial resistivity in ohm-cm.

        Notes
        -----
        Leaving default of None produces no change.
        """
        valid_sec_names = list(self.sections.keys())
        _check_option("sec_name", sec_name, valid_sec_names)

        if L is not None:
            _validate_type(L, (float, int), "L")
            self.sections[sec_name]._L = L

        if diam is not None:
            _validate_type(diam, (float, int), "diam")
            self.sections[sec_name]._diam = diam

        if cm is not None:
            _validate_type(cm, (float, int), "cm")
            self.sections[sec_name]._cm = cm

        if Ra is not None:
            _validate_type(Ra, (float, int), "Ra")
            self.sections[sec_name]._Ra = Ra

        self._update_end_pts()



"""
==========================
03. Simulate Gamma Rhythms
==========================

This example demonstrates how to simulate gamma rhythms via the well
established pyramidal-interneuron-gamma mechanisms [1]_,
as detailed in the `HNN GUI gamma tutorial
<https://jonescompneurolab.github.io/hnn-tutorials/gamma/gamma>`_,
using HNN-Core.

We recommend you first review the GUI tutorial.
The workflow below recreates weak gamma rhythms similar to Figures 4 and 5
of the GUI tutorial, and strong gamma rhythms similar to Figure 12 and 13
in the GUI tutorial.
"""

# Authors: Mainak Jas <mjas@mgh.harvard.edu>
#          Sam Neymotin <samnemo@gmail.com>
#          Christopher Bailey <bailey.cj@gmail.com>

# sphinx_gallery_thumbnail_number = 2

import os.path as op

###############################################################################
# Let us import hnn_core

import hnn_core
from hnn_core import simulate_dipole, read_params, jones_2009_model

hnn_core_root = op.dirname(hnn_core.__file__)

###############################################################################
# Read the parameter file and print the between-cell connectivity parameters.
# Note that these are different compared with the 'default' parameter set used
# in, e.g., :ref:`sphx_glr_auto_examples_workflows_plot_simulate_alpha.py`.
params_fname = op.join(hnn_core_root, 'param', 'gamma_L5weak_L2weak.json')
params = read_params(params_fname)
print(params['gbar_L*'])

###############################################################################
# We'll add a tonic Poisson-distributed excitation to pyramidal cells and
# simulate the dipole moment in a single trial (the default value used by
# ``simulate_dipole`` is ``n_trials=params['N_trials']``).

net = jones_2009_model(params)

weights_ampa = {'L2_pyramidal': 0.0008, 'L5_pyramidal': 0.0075}
synaptic_delays = {'L2_pyramidal': 0.1, 'L5_pyramidal': 1.0}
rate_constant = {'L2_pyramidal': 140.0, 'L5_pyramidal': 40.0}
net.add_poisson_drive(
    'poisson', rate_constant=rate_constant, weights_ampa=weights_ampa,
    location='proximal', synaptic_delays=synaptic_delays,
    event_seed=1349)

###############################################################################
dpls = simulate_dipole(net, tstop=250.)
scaling_factor = 30000
dpls = [dpl.scale(scaling_factor) for dpl in dpls]  # scale in place

###############################################################################
# Take a look at how different cell types respond to the exogenous drive. Note
# the periodic firing pattern of all cell types. While the basket cells fire
# relatively synchronously, the pyramidal cell populations display a more
# varied pattern, in which only a fraction of cells reach firing threshold.
net.cell_response.plot_spikes_raster()

###############################################################################
# To confirm that the periodicity observed in the firing patterns correspond to
# a population oscillation in the gamma-range, we can plot the time-frequency
# representation together with the signal. Note that the network requires some
# time to reach steady state. Hence, we omit the first 50 ms in our analysis.

tmin = 50
trial_idx = 0  # pick first trial

# plot dipole time course and time-frequency representation in same figure
import numpy as np
import matplotlib.pyplot as plt

fig, axes = plt.subplots(2, 1, sharex=True, figsize=(6, 6),
                         constrained_layout=True)

dpls[trial_idx].plot(tmin=tmin, ax=axes[0], show=False)

# Create an fixed-step tiling of frequencies from 20 to 100 Hz in steps of 1 Hz
freqs = np.arange(20., 100., 1.)
dpls[trial_idx].plot_tfr_morlet(freqs, n_cycles=7, tmin=tmin, ax=axes[1])

###############################################################################
# Now, let us try to re-run the simulation with a tonic bias
# applied to the L5 Pyramidal cells. Notice that the oscillation waveform is
# more regular, with less noise due to the fact that the tonic depolarization
# dominates over the influence of the Poisson drive. By default, a tonic bias
# is applied to the entire duration of the simulation.
net.add_tonic_bias(cell_type='L5_pyramidal', amplitude=6.)
dpls = simulate_dipole(net, tstop=250., n_trials=1)
dpls = [dpl.scale(scaling_factor) for dpl in dpls]  # scale in place

dpls[trial_idx].plot()

###############################################################################
# Notice that the Layer 5 pyramidal neurons now fire nearly synchronously,
# leading to a synchronous activation of the inhibitory basket neurons,
# resulting in a low-latency IPSP back onto the pyramidal cells. The duration
# of the IPSP is ~20 ms, after which the combined effect of the tonic bias and
# Poisson drive is to bring the pyramidal cells back to firing threshold,
# creating a ~50 Hz PING rhythm. This type of synchronous rhythm is sometimes
# referred to as “strong” PING.
net.cell_response.plot_spikes_raster()

###############################################################################
# Although the simulated dipole signal demonstrates clear periodicity, its
# frequency is lower compared with the "weak" PING simulation above.
from hnn_core.viz import plot_psd
plot_psd(dpls[trial_idx], fmin=20., fmax=100., tmin=tmin)

###############################################################################
# Finally, we demonstrate the mechanistic link between PING and the GABAA decay
# time constant (`tau2`). Using the same network/drive configuration as before,
# we decrease `tau2` from 5 to 2 ms. This will shorten the effective
# refactory period between L5 pyramidal cell spikes and increase the PING
# frequency from ~50 to ~65 Hz.
net.cell_types['L5_pyramidal']["cell_object"].synapses['gabaa']['tau2'] = 2
dpls = simulate_dipole(net, tstop=250., n_trials=1)
dpls = [dpl.scale(scaling_factor) for dpl in dpls]  # scale in place

fig, axes = plt.subplots(3, 1, sharex=True, figsize=(6, 6),
                         constrained_layout=True)
dpls[trial_idx].plot(ax=axes[0], show=False)
net.cell_response.plot_spikes_raster(ax=axes[1], show=False)
dpls[trial_idx].plot_tfr_morlet(freqs, n_cycles=7, tmin=tmin, ax=axes[2])

###############################################################################
# References
# ----------
# .. [1] Lee, S. & Jones, S. R. Distinguishing mechanisms of gamma frequency
#        oscillations in human current source signals using a computational
#        model of a laminar neocortical network. Frontiers in human
#        neuroscience (2013)



#! /usr/bin/env python
import platform
import os.path as op
import os
import subprocess
import shutil

from setuptools import setup, find_packages, Command
from setuptools.command.build_py import build_py

descr = """Code for biophysical simulation of a cortical column using Neuron"""

DISTNAME = 'hnn-core'
DESCRIPTION = descr
MAINTAINER = 'Mainak Jas'
MAINTAINER_EMAIL = 'mainakjas@gmail.com'
URL = ''
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'http://github.com/jonescompneurolab/hnn-core'

# get the version
version = None
with open(os.path.join('hnn_core', '__init__.py'), 'r') as fid:
    for line in (line.strip() for line in fid):
        if line.startswith('__version__'):
            version = line.split('=')[1].strip().strip('\"')
            break
if version is None:
    raise RuntimeError('Could not determine version')


# test install with:
# $ rm -rf hnn_core/mod/x86_64/
# $ rm -rf hnn_core/mod/arm64/
# $ python setup.py clean --all install
#
# to make sure there are no residual mod files
#
# also see following link to understand why build_py must be overridden:
# https://stackoverflow.com/questions/51243633/python-setuptools-setup-py-install-does-not-automatically-call-build
class BuildMod(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        print("=> Building mod files ...")

        if platform.system() == 'Windows':
            shell = True
        else:
            shell = False

        mod_path = op.join(op.dirname(__file__), 'hnn_core', 'mod')
        process = subprocess.Popen(['nrnivmodl'], cwd=mod_path,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, shell=shell)
        outs, errs = process.communicate()
        print(outs)


class build_py_mod(build_py):
    def run(self):
        self.run_command("build_mod")

        build_dir = op.join(self.build_lib, 'hnn_core', 'mod')
        mod_path = op.join(op.dirname(__file__), 'hnn_core', 'mod')
        shutil.copytree(mod_path, build_dir)

        build_py.run(self)


if __name__ == "__main__":
    extras = {
        'opt': ['scikit-learn'],
        'parallel': ['joblib', 'psutil'],
        'test': ['codespell', 'pytest', 'pytest-cov', 'pytest-xdist', 'ruff'],
        'docs': ['mne', 'myst-parser', 'nibabel', 'numpydoc', 'pillow',
                 'pooch', 'pydata-sphinx-theme', 'sphinx', 'sphinx-gallery',
                 'sphinx-copybutton', 'tdqm'],
        'gui': ['ipywidgets>=8.0.0', 'ipykernel', 'ipympl', 'voila'],
    }
    extras['dev'] = (extras['opt'] + extras['parallel'] + extras['test'] +
                     extras['docs'] + extras['gui'])


    setup(name=DISTNAME,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=version,
          download_url=DOWNLOAD_URL,
          long_description=open('README.md').read(),
          long_description_content_type='text/markdown',
          classifiers=[
              'Intended Audience :: Science/Research',
              'Intended Audience :: Developers',
              'License :: OSI Approved',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Microsoft :: Windows',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS',
              'Programming Language :: Python :: 3.9',
              'Programming Language :: Python :: 3.10',
              'Programming Language :: Python :: 3.11',
              'Programming Language :: Python :: 3.12',
              'Programming Language :: Python :: 3.13',
          ],
          platforms='any',
          install_requires=[
              'numpy >=1.14',
              'NEURON >=7.7, <=8.2.7; platform_system != "Windows"',
              'matplotlib>=3.5.3',
              'scipy',
              'h5io',
          ],
          extras_require=extras,
          python_requires='>=3.9, <3.14',
          packages=find_packages(),
          package_data={'hnn_core': [
              'param/*.json',
              'gui/*.ipynb']},
          cmdclass={'build_py': build_py_mod, 'build_mod': BuildMod},
          entry_points={'console_scripts': ['hnn-gui=hnn_core.gui.gui:launch']}
          )



"""
=================================
09. Replaying Spike Data as Input
=================================

Welcome to this hands-on tutorial on simulating communication between two neocortical
networks using HNN-Core! This notebook guides you through creating a source network
(Network A), extracting its spike data, feeding it into a target network (Network B)
using the `add_spike_train_drive` method, simulating both, and visualizing the
results. Let's dive in!
"""

# Authors: Maira Usman <maira.usman5703@gmail.com>

###############################################################################
# Step 1: Setting Up the Source Network (Network A)
# -------------------------------------------------
# 
# Our journey begins with Network A, the source of neural activity. We'll use the
# `jones_2009_model` to create a realistic neocortical network and simulate it with an
# evoked drive to generate spike data. This step mimics how a network might produce
# activity in a real brain region, which we'll later transmit to another network.
# 
# The simulation runs for 170 ms, and we'll save the spike data to a temporary file for
# efficient handling.
import numpy as np
import matplotlib.pyplot as plt
from hnn_core import jones_2009_model, simulate_dipole, read_spikes
import os.path as op
import tempfile

# Create and simulate Network A
print("Creating and simulating Network A...")
net_A = jones_2009_model()
net_A._params.update({'tstop': 170.0})

# Add a drive to generate activity
print("Adding drive to Network A...")
drive_params = {
    'name': 'evdist1', 'mu': 63.5, 'sigma': 3.8, 'numspikes': 1,
    'location': 'distal', 'event_seed': 274
}
weights = {
    'ampa': {'L2_basket': 0.006, 'L2_pyramidal': 0.0005, 'L5_pyramidal': 0.14},
    'nmda': {'L2_basket': 0.019, 'L2_pyramidal': 0.004, 'L5_pyramidal': 0.08}
}
delays = {'L2_basket': 0.1, 'L2_pyramidal': 0.1, 'L5_pyramidal': 0.1}
net_A.add_evoked_drive(
    **drive_params, weights_ampa=weights['ampa'], weights_nmda=weights['nmda'],
    synaptic_delays=delays
)

# Simulate and save spike data
print("Simulating Network A and saving spikes...")
dpls_A = simulate_dipole(net_A, tstop=170.0, n_trials=1)
with tempfile.TemporaryDirectory() as tmp_dir:
    spike_file = op.join(tmp_dir, 'spk_%d.txt')
    net_A.cell_response.write(spike_file)
    cell_response = read_spikes(op.join(tmp_dir, 'spk_*.txt'))
    print(f"Extracted {sum(len(t) for t in cell_response.spike_times)} spikes from Network A")

###############################################################################
# Step 2: Configuring the Target Network (Network B)
# --------------------------------------------------
# 
# Now, let's set up Network B, the target network that will receive spike data from
# Network A. We'll extract spikes from the first trial of Network A, filter for
# pyramidal cells, and format them into a dictionary compatible with
# `add_spike_train_drive`. This process is akin to distributing computed data across
# nodes in an MPI setup, where each node (here, Network B) processes a subset of the
# input.
# 
# We'll also define connection properties to map Network A's pyramidal cells to Network
# B's layers, ensuring a realistic interaction.
print("Creating and configuring Network B...")
net_B = jones_2009_model()
net_B._params.update({'tstop': 225.0})

# Extract spike data from first trial and filter for pyramidal cells
trial_idx = 0
spike_times = cell_response.spike_times[trial_idx]
spike_gids = cell_response.spike_gids[trial_idx]
spike_types = cell_response.spike_types[trial_idx]
pyramidal_mask = np.array([t in ['L2_pyramidal', 'L5_pyramidal'] for t in spike_types])
filtered_times = np.array(spike_times)[pyramidal_mask]
filtered_gids = np.array(spike_gids)[pyramidal_mask]
filtered_types = np.array(spike_types)[pyramidal_mask]
spike_data = {f"NetA_{t}_GID{g}": [] for t, g in zip(filtered_types, filtered_gids)}
for t, g, time in zip(filtered_types, filtered_gids, filtered_times):
    src_id = f"NetA_{t}_GID{g}"
    spike_data[src_id].append(time)

print(f"Number of valid pyramidal cells with spikes to be used for input: {len(spike_data)}")

# Define target configuration
conn_properties = {
    'L5_pyramidal': {'location': 'distal', 'weights_ampa': 0.005, 'synaptic_delays': 1.5},
    'L2_pyramidal': {'location': 'proximal', 'weights_ampa': 0.003, 'synaptic_delays': 1.0}
}
target_config = {}
source_ids = list(spike_data.keys())
for i, src_id in enumerate(source_ids):
    target_type = 'L5_pyramidal' if i % 2 == 0 else 'L2_pyramidal'
    props = conn_properties[target_type]
    target_config[src_id] = {
        'target_cell_types': [target_type],
        'location': props['location'],
        'weights_ampa': {target_type: props['weights_ampa']},
        'synaptic_delays': {target_type: props['synaptic_delays']},
        'probability': 0.7
    }
all_weights_ampa = {k: cfg['weights_ampa'][k] for cfg in target_config.values() for k in cfg['weights_ampa']}

# Add spike train drive to Network B
net_B.add_spike_train_drive(
    name='drive_from_NetA',
    spike_data=spike_data,
    location='distal',
    weights_ampa=all_weights_ampa,
    weights_nmda=None,
    synaptic_delays=0.1,
    conn_seed=42
)

###############################################################################
# Step 3: Simulating and Visualizing the Networks
# -----------------------------------------------
# 
# With Network B configured, we'll simulate it for 225 ms to observe the effect of
# Network A's spikes. Visualization is key in neuroscience, so we'll create raster plots
# for both networks. This step lets you see how the input from Network A influences
# Network B's activity.
print("Simulating Network B...")
dpls_b = simulate_dipole(net_B, tstop=225.0, n_trials=1)

# Visualize results
fig, axes = plt.subplots(2, 1, figsize=(10, 12), sharex=True, constrained_layout=True)
# 1. Network A: Spike raster
net_A.cell_response.plot_spikes_raster(ax=axes[0], show=False)
axes[0].set_title('Network A: Spike Raster')
axes[0].set_ylabel('Cell ID')
# 2. Network B: Spike raster
net_B.cell_response.plot_spikes_raster(ax=axes[1], show=False)
axes[1].set_title('Network B: Spike Raster')
axes[1].set_ylabel('Cell ID')
axes[1].set_xlabel('Time (ms)')
plt.show()

###############################################################################
# Now that the simulation is finished, let's verify the drive setup and ensure the
# connection was established correctly.
print("\nVerifying configuration:")
drive_name = 'drive_from_NetA'
if drive_name in net_B.external_drives:
    drive = net_B.external_drives[drive_name]
    print(f"  Drive '{drive_name}' created with {drive['n_drive_cells']} cells")
    if drive_name in net_B.gid_ranges:
        print(f"  GIDs assigned: {net_B.gid_ranges[drive_name]}")
    drive_connections = sum(1 for conn in net_B.connectivity if conn['src_type'] == drive_name)
    print(f"  Number of connections from drive: {drive_connections}")
else:
    print(f"  ERROR: Drive '{drive_name}' not found")

###############################################################################
# Conclusion
# -----------------------------------------------
# 
# You've now successfully replayed spikes from one network into another network! The
# raster plots show how Network A's activity influences Network B, while the
# verification confirms the drive setup. This approach is scalable—similar to MPI's
# distributed computing—allowing you to experiment with multiple networks or adjust
# parameters like weights and delays to explore different dynamics. Try modifying the
# `weights_ampa` or `synaptic_delays` values in Step 2 to see how they affect the
# results!
